# Typing Modules
import typing as typ

# External Modules
import os
import ase # type: ignore
from pymatgen.analysis.defects.point_defects 		import Interstitial # type: ignore
from pymatgen.transformations.site_transformations 	import InsertSitesTransformation # type: ignore
from pymatgen.io.ase 								import AseAtomsAdaptor # type: ignore
from pymatgen.symmetry.analyzer 					import SpacegroupAnalyzer # type: ignore

# Internal Modules
from catalog.misc.atoms import mag_elems

################################################################################

covalent_radii = {symb:rad for symb,rad in zip(ase.data.chemical_symbols,ase.data.covalent_radii)}

def getInterstitials(ase_in     : ase.Atoms
                    ,inter      : str
                    ,spinpol    : bool
                    ) -> typ.Any:
    """Helpful doc string"""
    pmg_init	= AseAtomsAdaptor.get_structure(ase_in)
    pmg_init2 	= SpacegroupAnalyzer(pmg_init).get_conventional_standard_structure()

    interstitial = Interstitial(pmg_init2,None,covalent_radii) #accuracy=high breaks...
    os.system('cls' if os.name == 'nt' else 'clear')
    output = []
    for i,site in enumerate(interstitial.enumerate_defectsites()):
        coordination =  int(round(interstitial.get_defectsite_coordination_number(i)))
        mult 	= 0 # interstitial.get_defectsite_multiplicity(i) -- broken ???
        insert  = InsertSitesTransformation([inter],[site.coords],coords_are_cartesian=True)
        try:
            pmg_new = insert.apply_transformation(pmg_init2.copy())
            ase_new = AseAtomsAdaptor.get_atoms(pmg_new)
            strname = '_%s-%dfold'%(inter,coordination)

            if spinpol:
                new_magmoms = [3 if e in mag_elems else 0 for e in ase_new.get_chemical_symbols()]
                ase_new.set_initial_magnetic_moments(new_magmoms)

            output.append((ase_new,strname))
        except ValueError: pass #ValueError: New site is too close to an existing site!
    return output



#Typing imports
import typing as typ

#External imports
import os

#Internal Imports
from dbgen.main import load_jobs as l_j
from catalog.catalog_config import SHERLOCK2_USERNAME,SUNCAT_USERNAME, USER

"""
Features:
- Syncing suncat -> sherlock
- Loading database (from local machine)
To do:
- underlying file system periodic backup
"""

################################################################################
# Sync Clusters
# -------------------

def sync_suncat()->None:
    print('Syncing /nfs/slac/g/suncatfs/ksb/share/jobs to /scratch/users/ksb/share/suncat_jobs_copy/ ...')
    slac  = '%s@suncatls1.slac.stanford.edu:/nfs/slac/g/suncatfs/ksb/share/jobs/%s/'%(SUNCAT_USERNAME,USER)
    sher  = '/scratch/users/ksb/share/suncat_jobs_copy/jobs/%s/'%(USER)
    rsync = 'rsync -rtqvu --perms --chmod=777 --omit-dir-times %s %s  --delete'%(slac,sher)
    ssh   = 'ssh %s@login.sherlock.stanford.edu '%SHERLOCK2_USERNAME
    os.system(ssh + rsync)

def sync_nersc()->None:
    print('Syncing /global/cscratch1/sd/krisb/share/jobs to /scratch/users/ksb/share/nersc_jobs_copy/ ...')
    os.system('rsync -rtqvu --perms --chmod=777 --omit-dir-times krisb@cori.nersc.gov:/global/cscratch1/sd/krisb/share/jobs /scratch/users/ksb/share/nersc_jobs_copy/ --delete')

#####################
# Load Database
# -------------------
def load_jobs(parallel : bool = True)-> None:
    sync_suncat()
    l_j(parallel = parallel
        ,catalog = True
        ,only    = """
                     catalog

                     metadata
                     anytraj
                     relax_job
                     subjobs

                     calc
                     calc_other
                     kptden

                     traj
                     systype
                     substructs

                     pop_struct_catalog
                     pop_ads_catalog
                     pop_bulk_struct_catalog
                     pop_surf_struct_catalog
                     """)
def ads_triple()->None:
    l_j(only = "ads_triple")



# # Typing Modules
#
#
# #External Modules
# import json,subprocess,os,shutil,itertools,pickle,glob,time
# import fireworks # type: ignore
# #Internal Modules
# import catalog.datalog.db_utils     as db
# # import catalog.detail.detail        as dt
# # import catalog.datalog.scrape       as scrape
#
# from catalog.misc.sql_utils  import *
# from catalog.misc.utilities     import flatten,merge_dicts,replacer
# from catalog.misc.print_parse   import (ask,print_sleep,abbreviate_dict
#                                        ,read_on_sher)
# ##########################
#
# """
# Functions for maintaining, updating, and creating a database for DFT calculations
#
# ### Database Maintenance
# - backup
# - delete
# - duplicates
#
# ### Updating
# - sync_suncat
# - update_db
#
# ### Database startup/generation
# - load_new
# - reset_all
# - load_jobs
# """
#
# ######################
# # Environment Variables
# # --------------------
# user   = os.environ['USER']
# try:
#     lpad    = fireworks.LaunchPad.from_file(os.environ.get('LAUNCHPAD_YAML'))
# except:
#     lpad    = None
# dbroot = '/scratch/users/ksb/share/'
# realdb = dbroot + 'suncatdata.db'
# tempdb = dbroot + 'temp.db'
#
# ######################
# # Database maintanence
# # --------------------
#
# def backup(force : bool = False) -> None:
#     root = dbroot+'backup/'
#     now  = time.time()
#
#     def getDaysSinceLastBackup(folderName : str) -> float:
#         return (now - os.path.getmtime(root+folderName+'/jobs'))/(24*3600)
#
#     def makeBackup(folderName : str) -> None :
#         print('backing up %s on sherlock'%folderName)
#         os.system('rsync  -rtqvu /scratch/users/ksb/share/jobs/ %s/jobs/ --delete'%(root+folderName))
#         os.system('touch -m %s/jobs'%(root+folderName)) #update mtime even if nothing transferred
#         print('backing up %s on suncat'%folderName)
#         suncout, err = subprocess.Popen(['ssh','{0}@suncatls1.slac.stanford.edu'.format(user), 'rsync  -rtqvu /nfs/slac/g/suncatfs/ksb/share/jobs/ /nfs/slac/g/suncatfs/ksb/share/backup/%s/jobs/ --delete'%folderName], stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
#         if user=='ksb':
#             print('backing up %s on nersc'%folderName)
#             nerscout, err = subprocess.Popen(['ssh','krisb@cori.nersc.gov', 'rsync  -rtqvu /global/cscratch1/sd/krisb/share/jobs/ /global/cscratch1/sd/krisb/share/backup/%s/jobs/ --delete'%folderName], stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
#             print(err)
#     lastMonth,lastWeek,lastDay = map(getDaysSinceLastBackup,['monthly','weekly','daily'])
#
#     if   lastMonth > 30: makeBackup('monthly')
#     elif lastWeek  > 7:  makeBackup('weekly')
#     elif lastDay   > 1 or force:
#         makeBackup('daily')
#         lpad.maintain(infinite=False)
#
# def delete(constraint : typ.Any
#           ,check      : bool    = True
#           ,archive    : bool    = True
#           ,deletePath : bool    = True
#           ) -> None:
#     """
#     Delete completed jobs that meet some constraint. Removes storage directory (but not working directory). Never delete from Temp.
#     """
#     # raise NotImplementedError
#     additionalConstraint = [USER_(user),job.deleted == 0]
#
#     output = db.Query([JOB,FWID,STORDIR,WORKDIR]
#                     ,AND(*(constraint + additionalConstraint))).query()
#     for jid,fwid,path,wpath in output:
#         if not check or ask('Do you want to delete %s?'%path):                          # Delete storage directory
#             if archive:
#                 try: lpad.archive_wf(fwid)                                                # Archive firework
#                 except ValueError: print('Could not archive fwid %d'%fwid)
#             db.updateDB('deleted',          'id',jid,1,                table='job')  # Note deletion in deleted column
#             db.updateDB('working_directory','id',jid,'formerly: '+wpath.replace('formerly: ',''),table='job')  # Modify due to UNIQUE constraint in schema
#             if deletePath:
#                 sherlock_trash = '/scratch/users/ksb/share/trash/'
#                 nfs_trash      = '/nfs/slac/g/suncatfs/ksb/share/trash/'
#                 nersc_trash    = '/global/cscratch1/sd/krisb/share/trash/'
#                 if 'scratch' in path:
#                     new_path = sherlock_trash + '%s/%s/'%(user,os.path.basename(path[1:len(path)-1]))
#                     shutil.move(path,new_path)
#                     print('moved from %s to %s'%(path, new_path))
#                 elif 'nfs' in path:
#                     new_path = nfs_trash + '%s/%s/'%(user,os.path.basename(path[1:len(path)-1]))
#                     d   = subprocess.Popen(['ssh','{0}@suncatls1.slac.stanford.edu'.format(user),
#                         'mv {0} {1}'.format(path,new_path)],
#                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
#                     dout, err = d.communicate()
#                     print('moved from %s to %s'%(path, new_path))
#                 elif 'global' in path:
#                     new_path = nersc_trash + 'krisb/%s/'%(os.path.basename(path[1:len(path)-1]))
#                     d   = subprocess.Popen(['ssh','krisb@cori.nersc.gov', 'mv {0} {1}'.format(path,new_path)],
#                                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
#                     dout, err = d.communicate()
#                     print('moved from %s to %s'%(path, new_path))
#                 else: raise NotImplementedError(path)
#                 db.updateDB('storage_directory','id',jid,new_path,table='job')  # Note deletion in deleted column <--- WHAT DOES THIS DO?
#             print('deleted!')
#
#
# #####################
# # Updating DB Columns
# # -------------------
#
# def sync_suncat()->None:
#     print('Syncing /nfs/slac/g/suncatfs/ksb/share/jobs to /scratch/users/ksb/share/suncat_jobs_copy/ ...')
#     os.system('rsync -rtqvu --perms --chmod=777 --omit-dir-times %s@suncatls1.slac.stanford.edu:/nfs/slac/g/suncatfs/ksb/share/jobs /scratch/users/ksb/share/suncat_jobs_copy/ --delete'%user)
# def sync_nersc()->None:
#     print('Syncing /global/cscratch1/sd/krisb/share/jobs to /scratch/users/ksb/share/nersc_jobs_copy/ ...')
#     os.system('rsync -rtqvu --perms --chmod=777 --omit-dir-times krisb@cori.nersc.gov:/global/cscratch1/sd/krisb/share/jobs /scratch/users/ksb/share/nersc_jobs_copy/ --delete')
#
#
#
#
# def update_db(cols       : typ.Optional[str]    = None
#              ,from_col   : typ.Optional[str]    = None
#              ,to_col     : typ.Optional[str]    = None
#              ,impt       : bool                 = True
#              ,retry      : bool                 = False
#              ,verbose    : bool                 = True
#              ,load       : bool                 = True
#              ,db_path    : str                  = realdb
#              ) -> None:
#     """
#     cols :: String
#         - Space separated list of column names
#         - Exclusive list of details to compute
#     from_col,to_col :: String
#         - Strings which denote the start and end columns for a range to details
#     impt :: Bool
#         - If true, don't compute useless/expensive things, like voronoi
#     retry :: Bool
#         - Overwrite existing values of columns
#     load :: Bool
#         - trigger load_db to be run
#     """
#
#     if cols is None: deets = dt.get_details(from_col,to_col,impt)
#     else:            deets = [dt.detail_dict[x] for x in cols.split()]
#
#     if load: load_jobs(db_path=db_path)
#
#     print('updating: <<%s>>'%','.join(map(str,deets)))
#
#     for d in deets:
#         d.apply(db_path,retry,verbose)
#         if str(d)=='refjob': fill_refeng(db_path=db_path)
#
# def fill_refeng(db_path : str = realdb) -> None:
#     """
#     Makes sure every relevant calculator has a row for each element
#     in refeng and every atoms has the right elements in composition
#
#     THIS SHOULD BE CALLED AFTER REFJOB IS APPLIED
#     """
#     cmd = 'INSERT OR IGNORE INTO refeng (calc_id,element_id,kptden_x,kptden_y) VALUES (?,?,?,?)'
#
#     binds = db.Query(cols     = [job.calc_id,REFJOB,KPTDEN_X,KPTDEN_Y]
#                     ,table    = job
#                     ,db_path  = db_path).query()
#
#     db.sqlexecutemany(cmd,binds,db_path=db_path) # type: ignore
#
# def fill_comp(db_path : str = realdb)->None:
#     """
#     Adds rows for each chemical composition
#
#     THIS SHOULD BE CALLED AFTER NEW JOBS ARE LOADED
#     """
#     cmd = 'INSERT OR REPLACE INTO composition (atoms_id,element_id,has) VALUES (?,?,?)'
#     atoms_elems = db.Query(cols  = [ATOMSID,NUMBER]
#                         ,table = atom
#                         ,distinct= True
#                         ,db_path=db_path).query()
#
#     binds = [(x,y,1) for (x,y) in atoms_elems]
#
#     db.sqlexecutemany(cmd,binds,db_path=db_path) # type: ignore
#
# #############################
# # Database startup/generation
# # ---------------------------
#
# def load_new(db_path : str = realdb) -> None:
#     if user=='ksb': backup()
#     load_jobs(db_path=db_path)
#     update_db(load=False,db_path=db_path)
#
# def reset_derived(db_path : str = realdb) -> None:
#     for d in dt.details: d.reset(db_path)
#
# def reset_from(from_col : str
#               ,db_path  : str = realdb
#               ) -> None:
#     for d in dt.get_details(from_col=from_col): d.reset(db_path)
#     update_db(from_col=from_col,load=True,retry=True,db_path=db_path)
#
# def reset_refeng(db_path : str = realdb) -> None:
#     db.drop_table('refeng',db_path=realdb)
#     db.add_refeng(db_path)
#     reset_from('refjob')
#
# def reset_all(limit     : typ.Optional[int] = None
#              ,update    : bool              = False
#              ,db_path   : str               = tempdb
#              ) -> None:
#     """
#     Repopulate completed table using the data stored in .../ksb/shared/jobs/...
#     """
#     db.removeSuncatDataDB(db_path) # removes temp
#     db.createSuncatDataDB(db_path) # creates temp
#     load_jobs(limit,db_path)       # loads temp
#     fill_comp(db_path)             # populates composition table (derived from atom table)
#     if update: update_db(load=False,db_path=db_path) # takes 2-3x longer
#
#
# def load_jobs(limit   : typ.Optional[int] = None
#              ,db_path : str =realdb
#              ) -> None:
#     """
#     Loads all completed jobs into the database
#     Scrapes any folder not in 'storage_directory' column for info in atoms/calc/job tables
#     """
#     sync_suncat() # Make sure .../share/suncatdataTEMP/ matches /nfs/slac/g/suncatfs/ksb/share/jobs
#     if user=='ksb': sync_nersc()
#     print("Loading recently completed jobs ...")
#
#     sharedjobsSherlock = '/scratch/users/ksb/share/jobs'
#     sharedjobsSuncat   = '/nfs/slac/g/suncatfs/ksb/share/jobs'
#     sharedjobsNersc    = '/global/cscratch1/sd/krisb/share/jobs'
#
#     suncout, serr = subprocess.Popen(['ssh','{0}@suncatls1.slac.stanford.edu'.format(user), 'cd %s;ls -d $PWD/*/*'%sharedjobsSuncat]
#                                     , stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate() # List of jobs in suncat stored jobs
#     if user == 'ksb':
#         nerscout, nerr = subprocess.Popen(['ssh','krisb@cori.nersc.gov', 'cd %s;ls -d $PWD/*/*'%sharedjobsNersc]
#                                     , stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate() # List of jobs in suncat stored jobs
#     else: nerscout = b''
#     sherout = os.popen('cd %s; ls -d $PWD/*/*'%sharedjobsSherlock).read() # List of jobs in sherlock stored jobs
#     nerscout = nerscout.decode(); suncout = suncout.decode() #convert from bytes to str
#     all_jobs_raw = itertools.chain.from_iterable([x.split('\n') for x in [sherout,suncout,nerscout]])
#
#     allJobs = [x+'/' for x in all_jobs_raw if len(x)>0]              # Merge lists, filter any '' items
#     newJobs = set(allJobs)-set(db.Query(table=job,db_path=db_path).query_col(job.storage_directory)) # Identify jobs not present already in database
#
#     if limit is not None: newJobs = list(newJobs)[:limit] # type: ignore
#     if newJobs:
#         print('adding %d new jobs'%len(newJobs))
#
#         for i,path in enumerate(newJobs):
#             if i%20: print(20-i%20)
#             else: print_sleep(1)
#             load_folder(path,db_path=db_path)
#
#         fill_comp(db_path=db_path)
#
#     else: print('No new jobs to add')
#
#
# ###################
# # Guessing Job Type
# # -----------------
# def guessJobType(vibpckls:list,qnlog:str) -> str:
#     """
#     Decide whether job is a vib job, a collection of singlepoints (e.g. bulkmod/latticeopt), or a relax
#     """
#     if len(vibpckls) > 0:                                  return 'vib'          # for vib, read INITIAL atoms from any .traj file (final doesn't matter)
#     else:
#         bfgs = qnlog.count('BFGS:')                                              # for bulkmod or lattice opt, read INITIAL atoms from any .traj file (accuracy here matters for bulkmod, not latticeopt)
#         if bfgs > 1 and bfgs==qnlog.count('BFGS:    0'):   return 'singlepoints' # possibly relax converges with single point, exclude these jobs (bulkmods and latticeopts take more than one singlepoint)
#         else:                                              return 'relax'        # includes vc-relax
#
# class Directory(object):
#     def __init__(self
#                 ,log    : str
#                 ,qnlog  : str
#                 ,pwinp  : str
#                 ,incar  : str
#                 ,kptcar  : str
#                 ,poscar  : str
#                 ,potcar  : str
#                 ,outcar  : str
#                 ,jsons  : list
#                 ,trajs  : list
#                 ,vibpckls  : list
#                 ) -> None:
#         self.log      = log
#         self.qnlog    = qnlog
#         self.pwinp    = pwinp
#         self.incar    = incar
#         self.kptcar   = kptcar
#         self.poscar   = poscar
#         self.potcar   = potcar
#         self.outcar   = outcar
#         self.jsons    = jsons
#         self.trajs    = trajs
#         self.vibpckls = vibpckls
#         self.jobtype  = guessJobType(vibpckls,qnlog)
#
# def get_files(storpth : str) -> Directory:
#     """
#     This function condenses most of the I/O into one step.
#         Grabs content of log file + pw.inp (if it exists)
#         Grabs the name of anything with a traj or json extension
#         Identifies whether or not there are vib.pckl files
#     """
#
#     storpth = replacer(storpth,{'/nfs/slac/g/suncatfs/ksb/share':'/scratch/users/ksb/share/suncat_jobs_copy'
#                                ,'/global/cscratch1/sd/krisb/share':'/scratch/users/ksb/share/nersc_jobs_copy'})
#
#     def get(x:str)->str:
#         try: return read_on_sher(storpth+x)
#         except Exception as e: return ''
#     def getGlob(x:str)->list: return glob.glob(storpth+x)
#     knownFiles = [get(x) for x in ['log','qn.log','pw.inp','INCAR','KPOINTS','POSCAR','POTCAR','OUTCAR']]
#     manyFiles  = map(getGlob,['*.json','*.traj','vib*.pckl'])
#     filtered   = [x for x in manyFiles if 'FW.json' not in x]
#
#     return Directory(*list(knownFiles + filtered)) # type: ignore
#
# def load_folder(storpath : str ,db_path : str = realdb)->None:
#     """
#     INPUT: a json of runtime data (generated by log()). OUTPUT: populates a new row of completed table
#     """
#     print('loading ',storpath)
#
#     direc   = get_files(storpath)
#
#     iID    = scrape.get_atoms_id(direc,True,db_path=db_path) # Assign IDs to init/final atoms. Add to DB if doesn't exist
#     fID    = scrape.get_atoms_id(direc,False,db_path=db_path) #
#     calcid = scrape.get_calc_id(direc,db_path=db_path)           # Assign IDs to calc params. Add to DB if doesn't exist
#
#     job_name = scrape.get_job_name(direc)
#     rawruntime = json.loads(read_on_sher(storpath+'runtime.json'))    # Runtime information
#     kwargs     = json.loads(rawruntime.pop('kwargs'))                    # Take kwargs and move to seperate object
#
#
#     if job_name is None:
#         job_name   = kwargs.get('job_name','')
#
#     runtime    = merge_dicts([rawruntime,{'calc_id': calcid
#                                         ,'initatoms':iID
#                                         ,'finalatoms':fID
#                                         ,'guess_jobtype':direc.jobtype
#                                         ,'job_name':job_name}])                                               # Look at *calc* table and get ID (if no matching calc exists, add a row to table)
#
#     colNames,binds = ','.join(runtime.keys()), list(runtime.values())                            # Runtime column names and corresponding values
#     command = 'INSERT into job (%s) VALUES (%s) '%( colNames, ','.join(['?']*len(binds)) ) # Make SQL command and
#     db.sqlexecute(command,binds,db_path=db_path) # type: ignore



#External imports
import sys
from glob import glob
import venv # type: ignore
import os, shutil

#Check to confirm that python3.6 or newer is being used
major_version, minor_version = sys.version_info[:2]
if major_version < 3 or minor_version < 6:
    raise Exception("Python 3.6 or a more recent version is required.")

class UserDefinedVariable:
    def __init__(self
                ,name
                ,desc       = ''
                ,domain     = None
                ,test       = lambda x: True
                ,process    = lambda x: x):
        """
        Class for storing information on user defined variables

        Parameters
        ----------
        name   : str
            The name the variable will be stored as.
        desc   : str
            A description of the user defined variable to be displayed
        domain : list
            A list of values that are allowed for the user defined variables.
            Defaults to None which allows all values
        test   : function
            A function that will be applied to the value given. Defaults to
            store_true
        process: function
            A function to process the user input (i.e. turn a path into an absolute path
            with os.path.abspath)
        """
        self.name    = name
        self.desc    = desc
        self.domain  = domain
        self.test    = test
        self.process = process

    def get_user_val(self):
        print('#####################')
        print(self.name+': '+self.desc)
        user_value = None
        while user_value == None:
            user_value = input("Please Enter your value for {}: ".format(self.name))
            if not self.test(user_value):
                user_value = None
            else:
                if (not self.domain is None and not user_value in self.domain):
                    user_value = None
                    print("Please enter a valid {} from the following domain:\n{}".format(self.name,self.domain))
        self.user_value = self.process(user_value)
        os.environ[self.name] = user_value
        print("value for {} set to {}".format(self.name,self.user_value))
        print('#-------------------')
        return self.user_value

def build_venv(env_dir):
    #Make the Virtual enviroment folder
    safe_mkdir(env_dir)
    if os.path.exists(os.path.join(env_dir,'.env')):
        print("""
        !!!!!!WARNING!!!!!!
        This folder already contains a .env directory
        This will be deleted if you continue
        """)
        answer = ''
        while not answer.lower() in ['n','y']:
            answer = input('Would you like to continue (y/n)? ')
        if answer.lower() =='n':
            print('not rebuilding the venv')
        else:
            venv.create(os.path.join(env_dir,'.env'),with_pip = True,symlinks = True, clear = True)
    else:
        #Build the Virtual enviroment and set the required user variables
        venv.create(os.path.join(env_dir,'.env'),with_pip = True,symlinks = True, clear = True)


    shell = os.getenv('SHELL','')
    rcpath = os.path.join(env_dir,'.env/bin/activate')
    user_vars_dict = get_user_enviroment_vars()
    user_vars_dict['CATALOG_LOC'] = env_dir
    set_user_enviromental_vars(rcpath,        user_vars_dict)
    set_user_enviromental_vars(rcpath+'.csh', user_vars_dict)

    #Pip install the required packages
    python_exec = os.path.join(env_dir,'.env/bin/python')
    os.system("{} -m pip install pip --upgrade --no-cache-dir".format(python_exec))
    os.system("{} -m pip install numpy --upgrade --no-cache-dir".format(python_exec))
    os.system("{} -m pip install CC-CataLog --upgrade --no-cache-dir".format(python_exec))

    # #Add the fireworks folder and the required sub-directories
    add_fireworks_folders(user_vars_dict)

    #Get the user to set the $CATALOG_LOC variable in their bashrc
    if 'bash' in shell:
        print("""
        !!!!!!IMPORTANT!!!!!
        NEED TO DO ONE MORE THING
        PLEASE RUN THIS COMMAND
        echo export CATALOG_LOC={}>>~/.bashrc

        To Source this enviroment in the future use the command:
        source $CATALOG_LOC/.env/activate
        or you can alias it:
        echo 'alias cata_src="source $CATALOG_LOC/.env/bin/activate"' >> ~/.bash_profile
        """.format(env_dir))
    elif 'tcsh' in shell:
        print("""
        !!!!!!IMPORTANT!!!!!
        NEED TO DO ONE MORE THING
        PLEASE RUN THIS COMMAND
        echo setenv CATALOG_LOC {}>>~/.cshrc

        To Source this enviroment in the future use the command:
        source $CATALOG_LOC/.env/activate.csh
        or you can alias it:
        echo 'alias cata_src "source $CATALOG_LOC/.env/bin/activate.csh"' >> ~/.cshrc
        """.format(env_dir))

def get_user_enviroment_vars():
    #Set of required user-defined variables
    def file_exist_test(file_curr):
        if os.path.exists(file_curr):
            return True
        else:
            print("{} does not exist".format(file_curr))
            return False
    user_vars = [UserDefinedVariable(name = "USER"
                                    ,desc = "Generic username used as default for new clusters"
                                    )
                ,UserDefinedVariable(name = "SHERLOCK2_USERNAME"
                                    ,desc = "Username for Sherlock2 cluster"
                                    ,)
                ,UserDefinedVariable(name = "SUNCAT_USERNAME"
                                    ,desc = "Username for Suncat Cluster"
                                    )
                ,UserDefinedVariable(name   = "HOSTNAME"
                                    ,desc   = "Name of current cluster (sherlock, suncat, local"
                                    ,domain = ['sherlock','local','suncat']
                                    )
                ,UserDefinedVariable(name   = "LAUNCHPAD_YAML"
                                    ,desc   = "Path to your my_launchpad.yaml file, relative paths are fine\nIf you don't have a fireworks database yet please get one."
                                    ,test   = file_exist_test
                                    ,process = lambda file_curr: os.path.abspath(file_curr)
                                    )
                ,UserDefinedVariable(name   = "DB_JSON"
                                    ,desc   = "Path to your SQL database credentials json."
                                    ,test   = file_exist_test
                                    ,process = lambda file_curr: os.path.abspath(file_curr)
                                    )
                ]
    user_vars_dict = {} # type: dict
    #Print a description of each variable and ask the user for input
    for user_var in user_vars:
        user_vars_dict[user_var.name] = user_var.get_user_val()
    return user_vars_dict

def set_user_enviromental_vars(rcpath,user_vars_dict):
    with open(rcpath,'a') as f:
        f.write("#####################\n")
        f.write("#Catalog enviromental variables\n")
        f.write("#####################\n")
        if rcpath.endswith('csh'):
            f.write("setenv PYTHONPATH {}:$PYTHONPATH\n".format(user_vars_dict['CATALOG_LOC']))
            for key,val in user_vars_dict.items():
                f.write("setenv {} {}\n".format(key,val))
        else:
            f.write("export PYTHONPATH={}:$PYTHONPATH\n".format(user_vars_dict['CATALOG_LOC']))
            for key,val in user_vars_dict.items():
                f.write("export {}={}\n".format(key,val))

def add_fireworks_folders(user_vars_dict):
    #Make all the necessary folders
    safe_mkdir(os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks'))
    safe_mkdir(os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks','jobs'))
    safe_mkdir(os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks','FW_CONFIG_DIR'))
    safe_mkdir(os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks','logs'))

    #Make a FW_config.yaml
    with open(os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks','FW_config.yaml'),'w') as f:
        f.write('ALWAYS_CREATE_NEW_BLOCK: True')

    #Copy the DB_JSON file
    shutil.copy(user_vars_dict['DB_JSON'],os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks','DB.json'))

    #Fill the FW_CONFIG_DIR with my_launchpad, my_fworker, and my_qadapter
    #Copy the launchpad_yaml
    shutil.copy(user_vars_dict['LAUNCHPAD_YAML'],os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks','FW_CONFIG_DIR','my_launchpad.yaml'))


    #Write the my_fworker.yaml
    with open(os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks','FW_CONFIG_DIR','my_fworker.yaml'),'w') as f:
        f.write("name: {}\n".format(user_vars_dict['HOSTNAME'].replace('sherlock','sherlock2')))
        f.write("category: ''\n")
        f.write("query: '{}'\n")

    if not user_vars_dict['HOSTNAME'] == 'local':
        #Write the my_qadapter.yaml
        #Get the queueing system from the user
        queue_system = None
        print('#####################')
        while not queue_system in ['SLURM','LSF','PBS','None']:
            queue_system = input("Which queueing system does this cluster use (SLURM/LSF/PBS/None)?")
        if queue_system == 'None':
            print("Since the queueing system was set to None, this machine can not be used to run jobs\n It can still launch them though!")
        print('#-------------------')

        template_dict = {'LSF'   : 'LoadSharingFacility_template.txt'
                        ,'PBS'   : 'PBS_template.txt'
                        ,'SLURM' : 'SLURM_template.txt'
                        ,'None'  : 'None'}
        glob_template = glob('{}/.env/lib/*/site-packages/catalog/fw/qadapter_templates/{}'.format(user_vars_dict['CATALOG_LOC'],template_dict[queue_system]))
        if len(glob_template)>0:
            template_file = glob_template[0]
        else:
            template_file = 'null'
            print('Template File not found please ask Michael Statt to set your template file!!!!')
        qtype_dict    = {'LSF'   : 'LoadSharingFacility'
                        ,'PBS'   : 'PBS'
                        ,'SLURM' : 'SLURM'
                        ,'None'  : 'null'}

        with open(os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks','FW_CONFIG_DIR','my_qadapter.yaml'),'w') as f:
            f.write("_fw_name: CommonAdapter\n")
            f.write("_fw_q_type: {}\n".format(qtype_dict[queue_system]))
            if not template_file == '':
                f.write("_fw_template_file: {}\n".format(template_file))
            f.write("queue: {}\n".format(user_vars_dict['HOSTNAME'].replace('sherlock2','iric')))
            f.write("rocket_launch: rlaunch -c $CATALOG_LOC/FW_CONFIG_DIR singleshot\n")

        #Copy the launcher function to the fireworks folder
        with open(os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks','launcher.sh'),'w') as f:
            f.write("#!/bin/bash\n")
            f.write("cd $CATALOG_LOC/fireworks\n")
            f.write("source $CATALOG_LOC/.env/bin/activate\n")
            f.write("$CATALOG_LOC/.env/bin/qlaunch -c $CATALOG_LOC/fireworks/FW_CONFIG_DIR --logdir $CATALOG_LOC/fireworks/logs -r rapidfire\n")
        os.chmod(os.path.join(user_vars_dict['CATALOG_LOC'],'fireworks','launcher.sh'),0o755)

def safe_mkdir(dir_path):
    if not os.path.exists(dir_path):
        os.mkdir(dir_path)

if __name__ == '__main__':
    env_dir2 = input('Please enter the path you would like to put the CataLog enviroment\n(Please choose an empty or non-existent folder): ')
    env_dir2 = os.path.abspath(env_dir2)
    build_venv(env_dir2)



#Typing imports
import typing as typ

#External imports
from PyQt5 import QtCore, QtGui, QtWidgets # type: ignore
import qdarkstyle                          # type: ignore
import logging, inspect


#Internal Imports
import catalog.gendata.gendata as gendata
from catalog.datalog.db_utils import Query
from catalog.misc.sql_utils import *
from catalog.gui.plotly_viewer import PlotlyView
from catalog.jobs.cluster import cluster_dict
from adsorber import filter_functions as ff


class QPlainTextEditLogger(logging.Handler):
    def __init__(self, parent : typ.Any) -> None:
        super().__init__()
        self.widget = QtWidgets.QPlainTextEdit(parent)
        self.widget.setReadOnly(True)

    def emit(self, record : typ.Any)-> None:
        msg = self.format(record)
        self.widget.appendPlainText(msg)

class CheckableComboBox(QtWidgets.QComboBox):
    def __init__(self):
        super(CheckableComboBox, self).__init__()
        self.view().pressed.connect(self.handleItemPressed)
        self.setModel(QtWidgets.QStandardItemModel(self))

    def handleItemPressed(self, index):
        item = self.model().itemFromIndex(index)
        if item.checkState() == QtCore.Qt.Checked:
            item.setCheckState(QtCore.Qt.Unchecked)
        else:
            item.setCheckState(QtCore.Qt.Checked)

class GenDataForm(QtWidgets.QMainWindow):
    def __init__(self
                ,job_type : str
                ,job_ids   : typ.List[int]
                ,parent    : typ.Any      = None
                )->None:
        super(GenDataForm, self).__init__()
        self.neglected_args    = ['limit','constraints','check_func','verbose'] #Args that user can specify

        self.gendata_functions = {'Adsorb'          : gendata.adsorb
                                 ,'Bare'            : gendata.bare
                                 ,'Bare All Facets' : gendata.bare_all_facets
                                 ,'Vacancy'         : gendata.vacancy}

        self.parent     = parent
        self.job_ids    = job_ids
        self.job_type   = job_type
        self.atoms_objs = Query(constraints = [JOB_IN_(job_ids)]).make_atoms()  # type: ignore
        self.arg_dict   = self.get_arg_dict(self.gendata_functions[self.job_type])

        self.setupUi()
        self._connect_signals()
        # self._update_input_table()
        logging.info('Hello')

    def _connect_signals(self)->None:
        self.push_submit.clicked.connect(self._submit)
        self.push_cancel.clicked.connect(self._cancel)

    def _submit(self)-> None:
        constraints       = [JOB_IN_(self.job_ids)]
        limit             = len(self.job_ids)
        gendata_kwargs    = self._get_gendata_kwargs_from_table()
        gendata_kwargs.update({'constraints': constraints
                              ,'limit'      : limit
                              ,'verbose'    : False
                              ,'check_func' : lambda x: True})
        gendata_obj       = self.gendata_functions[self.job_type](**gendata_kwargs)
        submit_kwargs     = self._get_cluster_info()
        gendata_dict      = gendata_obj.submit(**submit_kwargs)
        self.parent.central_tabWidget.removeTab(1)
        self.parent.central_tabWidget.setCurrentIndex(0)
        self.submited_ids = list(gendata_dict.values())

    def _cancel(self) -> None:
        self.parent.central_tabWidget.removeTab(1)
        self.parent.central_tabWidget.setCurrentIndex(0)


    def _get_gendata_kwargs_from_table(self) -> dict:
        kwargs = {} # type: ignore
        for row_i,arg_name in enumerate(self.arg_dict.keys()):
            try:
                if self.arg_edits[row_i].text() in [None,'']:
                    kwargs[arg_name] = None
                else:
                    kwargs[arg_name] = eval(self.arg_edits[row_i].text())
            except NameError:
                logging.error('Arguement {} has invalid value {}'.format(arg_name,self.arg_edits[row_i].text()))
                break
        return kwargs

    def _get_cluster_info(self) -> dict:
        walltime    = self.walltime_spinbox.value()
        clust_str   = self.cluster_combobox.currentText()
        nodes       = int(self.nodes_spinbox.value())
        return {'clust_str':clust_str,'nodes':nodes,'walltime':walltime}

    def get_arg_dict(self, func : typ.Any) -> dict:
        """
        returns a dictionary of arg_name:default_values for the input function
        """
        argspec          = inspect.getfullargspec(func)
        func_args        = argspec.args
        func_defaults    = argspec.defaults
        if func_args == []:
            return {}
        default_dict = dict(zip(func_args[-len(func_defaults):], func_defaults))
        return {arg : default_dict.get(arg) for arg in func_args if not arg in self.neglected_args}

    def setupUi(self) -> None:
        #Build Central Widget and set its layout
        self.centralwidget  = QtWidgets.QWidget(self)
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")

        #Build top frame
        self.top_frame = QtWidgets.QFrame()
        self.top_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.top_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.top_frame.setObjectName("frame")
        self.top_frame_layout = QtWidgets.QHBoxLayout(self.top_frame)
        self.top_frame_layout.setObjectName("top_frame_layout")

        #Build Push Box
        ###################
        self.push_box = QtWidgets.QGroupBox(self.top_frame)
        self.push_box.setObjectName("push_box")
        self.push_box_layout = QtWidgets.QFormLayout(self.push_box)
        self.push_box_layout.setObjectName("push_box_layout")

        #Create push_submit button
        self.push_submit = QtWidgets.QPushButton(self.push_box)
        self.push_submit.setObjectName("push_submit")
        self.push_submit.setText("Submit")
        self.push_box_layout.setWidget(1, QtWidgets.QFormLayout.SpanningRole, self.push_submit)

        #Create push_submit button
        self.push_cancel = QtWidgets.QPushButton(self.push_box)
        self.push_cancel.setObjectName("push_cancel")
        self.push_cancel.setText("Cancel")
        self.push_box_layout.setWidget(2, QtWidgets.QFormLayout.SpanningRole, self.push_cancel)

        #Create Gendata label and comboBox
        self.gendata_label = QtWidgets.QLabel(self.push_box)
        self.gendata_label.setObjectName("gendata_label")
        self.gendata_label.setText("GenData Function")
        self.push_box_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.gendata_label)

        self.gendata_label = QtWidgets.QLabel(self.job_type)

        self.push_box_layout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.gendata_label)
        self.top_frame_layout.addWidget(self.push_box)

        #Create Cluster box
        self.cluster_box = QtWidgets.QGroupBox(self.top_frame)
        self.cluster_box.setObjectName("cluster_box")
        self.cluster_box.setTitle("Cluster Information")
        self.formLayout = QtWidgets.QFormLayout(self.cluster_box)
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout.setLabelAlignment(QtCore.Qt.AlignCenter)
        self.formLayout.setFormAlignment(QtCore.Qt.AlignHCenter|QtCore.Qt.AlignTop)
        self.formLayout.setObjectName("formLayout")

        #Walltime label and field
        self.walltime_label = QtWidgets.QLabel(self.cluster_box)
        self.walltime_label.setObjectName("walltime_label")
        self.walltime_label.setText("Walltime")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.walltime_label)
        self.walltime_spinbox = QtWidgets.QSpinBox(self.cluster_box)
        self.walltime_spinbox.setMinimum(1)
        self.walltime_spinbox.setMaximum(150)
        self.walltime_spinbox.setValue(40)
        self.walltime_spinbox.setObjectName("walltime_spinbox")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.walltime_spinbox)

        #Cluster label and Field
        self.cluster_label = QtWidgets.QLabel(self.cluster_box)
        self.cluster_label.setObjectName("cluster_label")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.cluster_label)
        self.cluster_combobox = QtWidgets.QComboBox(self.cluster_box)
        self.cluster_combobox.setObjectName("cluster_combobox")
        self.cluster_combobox.addItems(list(cluster_dict.keys()))

        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.cluster_combobox)

        #Nodes label and Field
        self.nodes_label = QtWidgets.QLabel(self.cluster_box)
        self.nodes_label.setObjectName("nodes_label")
        self.nodes_label.setText("Nodes")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.nodes_label)
        self.nodes_spinbox = QtWidgets.QSpinBox(self.cluster_box)
        self.nodes_spinbox.setObjectName("nodes_spinbox")
        self.nodes_spinbox.setMinimum(1)
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.nodes_spinbox)
        self.top_frame_layout.addWidget(self.cluster_box)

        #Create Message Box
        self.message_box = QtWidgets.QGroupBox(self.top_frame)
        self.message_box.setObjectName("message_box")
        self.message_box.setTitle("Messages")
        self.message_box_layout = QtWidgets.QHBoxLayout(self.message_box)
        self.message_box_layout.setObjectName("message_box_layout")

        #Setup logger
        self.logTextBox = QPlainTextEditLogger(self)
        # # You can format what is printed to text box
        self.logTextBox.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
        logging.getLogger().handlers = []
        logging.getLogger().addHandler(self.logTextBox)
        # You can control the logging level
        logging.getLogger().setLevel(logging.DEBUG)
        self.message_box_layout.addWidget(self.logTextBox.widget)
        self.top_frame_layout.addWidget(self.message_box)
        self.verticalLayout.addWidget(self.top_frame)

        #Create bottom box and table
        self.bottom_box = QtWidgets.QGroupBox(self)
        self.bottom_box.setObjectName("bottom_box")
        self.bottom_box.setTitle("Inputs")
        self.bottom_box_layout = QtWidgets.QHBoxLayout(self.bottom_box)
        self.bottom_box_layout.setObjectName("bottom_box_layout")

        self.arg_box = QtWidgets.QGroupBox(self.bottom_box)
        self.arg_box.setTitle("Input Arguments")
        self.arg_box_layout = QtWidgets.QFormLayout(self.arg_box)

        self.arg_edits = []
        for arg, default in self.arg_dict.items():
            if isinstance(default,str):
                default = '\''+default+'\''
            self.arg_edits.append(QtWidgets.QLineEdit(str(default)))
            self.arg_box_layout.addRow(QtWidgets.QLabel(arg), self.arg_edits[-1])
        self.bottom_box_layout.addWidget(self.arg_box)

        self.plotly_viewer = PlotlyView(self.atoms_objs)
        self.bottom_box_layout.addWidget(self.plotly_viewer)
        self.bottom_box_layout.setStretch(1,1)

        self.verticalLayout.addWidget(self.bottom_box)

        self.verticalLayout.setStretch(1,1)
        self.setCentralWidget(self.centralwidget)
        QtCore.QMetaObject.connectSlotsByName(self)




if __name__ == "__main__":
    import sys,os
    os.environ['QT_API']='pyqt5'
    sys.argv.append('--disable-web-security')
    app = QtWidgets.QApplication(sys.argv)
    app.setStyleSheet(qdarkstyle.load_stylesheet_from_environment(is_pyqtgraph=False))

    gui = GenDataForm(job_type = 'Adsorb',job_ids = [27457,27459])
    gui.showMaximized()
    sys.exit(app.exec_())



# Typing modules
import typing as typ

#External modules
import os,json,time,re
import numpy as np              # type: ignore
import pymatgen as pmg          # type: ignore
from pymatgen.analysis.adsorption import AdsorbateSiteFinder, reorient_z, patches, color_dict, get_rot, Path # type: ignore


"""
Printing and parsing functions


np_to_str
read_on_sher
digit
alpha
alpha_num_split
ask
s
print_sleep
abbreviate_dict
read_storage_json
parse_line
storagedir_to_pckl
print_time
"""
################################################################################

def print_time(floatHours : float) -> str:
    intHours = int(floatHours)
    return "%02d:%02d" % (intHours,(floatHours-intHours)*60)


def read_storage_json(file_path : str) -> str:
    """

    """
    if file_path[0] == '.':
        file_path = os.getcwd()+file_path.replace('.','',1)
    file_text = read_on_sher(file_path)
    return json.loads(file_text)

def parse_line(string : str
              ,substr : str
              ,index  : int =0
              ) -> str:
    """
    Returns the first line containing substring
    """
    return [line for line in string.split('\n') if substr in line][index]


def abbreviate_dict(d : dict)->str:
    strdict =  '\n'.join(['%s : %s'%(k,v if len(str(v))<100 else '...') for k,v in sorted(d.items()) if v is not None])
    nonelist= "\n\n'None' keys: "+','.join(sorted([k for k,v in d.items() if v is None]))
    return strdict+nonelist

def s(x : str) -> str:
    """
    """
    return "'"+str(x)+"'" if isinstance(x,str) else str(x)

def ask(x : str)->typ.Any:
    """
    """
    return input(x+'\n(y/n)---> ').lower() in ['y','yes']

def print_sleep(n : int) -> None:
    """
    """
    for i in reversed(range(n)):
        print("sleeping ...." + (" %d ..."%i if n > 1 else ''))
        time.sleep(1)


def np_to_str(np_obj : np.array
             ,n 	 : int  	= 3
             ,matrix : bool 	= True
             ) -> str:
    """
    The result should be JSON parsable back to a list (or list of lists)
    """
    def rnd(x : float) -> float:
        x = x if abs(x) > 10**(-n) else abs(x)
        return ('{:.%df}'%n).format(x,prec=n) # type: ignore
    if matrix: return str(map(lambda y: map(lambda x: rnd(x),y),np_obj)).replace("'",'')
    else:      return str(map(lambda x: rnd(x),np_obj)).replace("'",'')

def get_sher_pth(pth : str) -> str:
    if   'nfs'    in pth: pth = pth.replace('/nfs/slac/g/suncatfs/ksb/share',  '/scratch/users/ksb/share/suncat_jobs_copy')
    elif 'global' in pth: pth = pth.replace('/global/cscratch1/sd/krisb/share','/scratch/users/ksb/share/nersc_jobs_copy')
    return pth

def read_on_sher(pth : str) -> str:
    """
    Presumes we've rsync'd suncat/share/jobs to sherlock/share/suncat_jobs_copy
    """
    pth_on_sher = get_sher_pth(pth)
    with open(pth_on_sher,'r') as f: return f.read()


def digit(x : str)->str:    return re.sub("[^0-9]","",x)
def alpha(x : str)->str:    return re.sub("[^a-zA-Z]", "", x)
def alpha_num_split(x:str)->typ.Tuple[str,str]:
    return (alpha(x),digit(x))

def roundfloat(x : typ.Any) -> float:
    output =  round(float(x),3)
    return abs(output) if  output == 0 else output





###################
#Archived Scripts
###################
# def folder_name(path : st,updir=0):
#     import os
#     if updir==0:
#         return os.path.basename(path)
#     return os.path.basename(path[0:path.find('/'+folder_name(path, updir-1)[0])])
# def plot_slab(slab 			   : pmg.Structure
#              ,ax 			   : typ.Any
#              ,scale			   : float					= 0.8
#              ,repeat		   : int					= 5
#              ,window		   : float					= 1.5
#              ,draw_unit_cell   : bool   				= True
#              ,decay			   : float  				= 0.2
#              ,adsorption_sites : bool					= True
#              ,site_type 	   : str					= 'all'
#              ,symm_reduce	   : float					= 0.01
#              ,sites_to_draw    : typ.Optional[np.array]	= None
#              ,height 		   : float					= 0.9
#              ) -> typ.Any:
#     """Function that helps visualize the slab in a 2-D plot, for
#     convenient viewing of output of AdsorbateSiteFinder.
#
#     Args:
#         slab (slab): Slab object to be visualized
#         ax (axes): matplotlib axes with which to visualize
#         scale (float): radius scaling for sites
#         repeat (int): number of repeating unit cells to visualize
#         window (float): window for setting the axes limits, is essentially
#             a fraction of the unit cell limits
#         draw_unit_cell (bool): flag indicating whether or not to draw cell
#         decay (float): how the alpha-value decays along the z-axis"""
#
#     orig_slab = slab.copy()
#     # slab = reorient_z(slab)
#     orig_cell = slab.lattice.matrix.copy()
#     if repeat:
#         slab.make_supercell([repeat, repeat, 1])
#     coords = np.array(sorted(slab.cart_coords, key=lambda x: x[2]))
#     sites = sorted(slab.sites, key=lambda x: x.coords[2])
#     alphas = 1 - decay * (np.max(coords[:, 2]) - coords[:, 2])
#     alphas = alphas.clip(min=0)
#     corner = [0, 0, slab.lattice.get_fractional_coords(coords[-1])[-1]]
#     corner = slab.lattice.get_cartesian_coords(corner)[:2]
#     verts = orig_cell[:2, :2]
#     lattsum = verts[0] + verts[1]
#
#     # Draw circles at sites and stack them accordingly
#     for n, coord in enumerate(coords):
#         r = sites[n].specie.atomic_radius * scale
#         ax.add_patch(patches.Circle(coord[:2] - lattsum * (repeat // 2),
#                                     r, color='w', zorder=2 * n))
#         color = color_dict[sites[n].species_string]
#         ax.add_patch(patches.Circle(coord[:2] - lattsum * (repeat // 2), r,
#                                     facecolor=color, alpha=alphas[n],
#                                     edgecolor='k', lw=0.3, zorder=2 * n + 1))
#     # Adsorption sites
#     if adsorption_sites:
#         asf = AdsorbateSiteFinder(orig_slab, height=height)
#         ads_sites_to_plot = asf.find_adsorption_sites(symm_reduce = symm_reduce)[site_type]
#         sop = get_rot(orig_slab)
#         ads_sites_to_plot = [sop.operate(ads_site_to_plot)[:2].tolist()
#                      for ads_site_to_plot in ads_sites_to_plot]
#         ax.plot(*zip(*ads_sites_to_plot), color='k', marker='x',
#                 markersize=10, mew=1, linestyle='', zorder=10000)
#
#     if not sites_to_draw is None:
#         site_xycoords = map(lambda x: x.pos[:2], sites_to_draw)
#         ax.plot(*zip(*site_xycoords), color='k', marker='x',
#                 markersize=10, mew=1, linestyle='', zorder=10000, picker = 5)
#
#     # Draw unit cell
#     if draw_unit_cell:
#         verts = np.insert(verts, 1, lattsum, axis=0).tolist()
#         verts += [[0., 0.]]
#         verts = [[0., 0.]] + verts
#         codes = [Path.MOVETO, Path.LINETO, Path.LINETO,
#                  Path.LINETO, Path.CLOSEPOLY]
#         verts = [(np.array(vert) + corner).tolist() for vert in verts]
#         path = Path(verts, codes)
#         patch = patches.PathPatch(path, facecolor='none', lw=2,
#                                   alpha=0.5, zorder=2 * n + 2)
#         ax.add_patch(patch)
#     ax.set_aspect("equal")
#     center = corner + lattsum / 2.
#     extent = np.max(lattsum)
#     lim_array = [center - extent * window, center + extent * window]
#     x_lim = [ele[0] for ele in lim_array]
#     y_lim = [ele[1] for ele in lim_array]
#     ax.set_xlim(x_lim)
#     ax.set_ylim(y_lim)
#     return ax



#Typing imports
import typing as typ

#External imports
from PyQt5 import QtCore, QtGui, QtWidgets # type: ignore
import qdarkstyle                          # type: ignore
#Internal Imports
from catalog.gui.job_maker import JobMakerGUI
from catalog.gui.relauncher import  RelaunchGUI


class CataLogGUI(QtWidgets.QMainWindow):
    def __init__(self
                ,parent          : typ.Any       = None
                ,guis            : list          = []
                ,gui_names       : list          = ['Job Maker', "Relauncher"]
                ) -> None:
        QtWidgets.QWidget.__init__(self, parent)
        self.guis = guis
        self.gui_names = gui_names
        self.setupUi()

    def setupUi(self) -> None:
        """
        Create the UI for the CataLogGUI object
        """
        self.setObjectName("Query")
        self.resize(1900, 732)
        self.setWindowTitle( "Query GUI")

        #Set Fonts for Push Buttons and Titles
        self.header_font = QtGui.QFont()
        self.header_font.setFamily("Helvetica Neue")
        self.header_font.setPointSize(15)

        self.centralwidget = QtWidgets.QWidget(self)
        self.centralwidget.setFont(self.header_font)
        self.centralwidget.setObjectName("centralwidget")
        self.central_layout = QtWidgets.QHBoxLayout(self.centralwidget)
        self.central_layout.setObjectName("central_layout")
        self.central_tabWidget = QtWidgets.QTabWidget()
        self.central_layout.addWidget(self.central_tabWidget)

        self.tabs        = [] # type: list
        self.tab_layouts = [] # type: list
        for gui,name in zip(self.guis,self.gui_names):
            self.tabs.append(QtWidgets.QWidget())
            self.tab_layouts.append(QtWidgets.QHBoxLayout(self.tabs[len(self.tabs)-1]))
            self.tab_layouts[len(self.tabs)-1].addWidget(gui)
            self.central_tabWidget.addTab(self.tabs[len(self.tabs)-1],name)


        self.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(self)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1319, 22))
        self.menubar.setDefaultUp(True)
        self.menubar.setObjectName("menubar")
        self.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(self)
        self.statusbar.setObjectName("statusbar")
        self.setStatusBar(self.statusbar)

        self.central_tabWidget.setCurrentIndex(0)
        self.central_tabWidget.setDocumentMode(True)

def main() -> None:
    import sys,os
    os.environ['QT_API']='pyqt5'
    sys.argv.append('--disable-web-security')
    app = QtWidgets.QApplication(sys.argv)
    app.setStyleSheet(qdarkstyle.load_stylesheet_from_environment(is_pyqtgraph=False))
    query = """
SELECT job_id,
   job_name,
   USER,
   stordir
FROM job
JOIN relax_job USING (job_id)
JOIN calc USING (calc_id)
JOIN calc_other USING (calc_other_id)
JOIN finaltraj USING (job_id)
JOIN struct USING (struct_id)
LEFT JOIN surface USING (struct_id)
LEFT JOIN bulk USING (struct_id)
LEFT JOIN molecule USING (struct_id)
WHERE user = 'aayush' and cell_dofree != ''
LIMIT 100
    """
    guis = [JobMakerGUI(query=query),RelaunchGUI()]
    gui_names = ['Job Maker','Relauncher']
    gui = CataLogGUI(guis = guis, gui_names = gui_names)
    gui.showMaximized()
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



# External modules
import typing as typ

import numpy as np                                              # type: ignore
import ase, itertools                                           # type: ignore
import pymatgen as pmg                                          # type: ignore
from ase.constraints               import FixAtoms              # type: ignore
from pymatgen.symmetry.analyzer    import SpacegroupAnalyzer    # type: ignore
from pymatgen.io.ase               import AseAtomsAdaptor       # type: ignore
from pymatgen.core.surface         import (SlabGenerator,Slab,  # type: ignore
                                          get_symmetrically_distinct_miller_indices)    # type: ignore
from pymatgen.analysis.adsorption  import AdsorbateSiteFinder,reorient_z # type: ignore
from pymatgen.core.operations      import SymmOp                           # type: ignore

from catalog.misc.utilities import flatten
################################################################################

"""
Functions to allow a user to create surfaces from completed bulk jobs
"""

#####################################
# Functions for constructing surfaces
#####################################
def make_slab(unit                      : pmg.Structure
              ,facet                    : typ.Tuple[int,int,int]
              ,num_atoms_per_layer      : int
              ,vac                      : float
              ,layers                   : int
              ,sym                      : bool
              ) -> typ.List["ase.Atoms"]:
    """
    Finds the slab thickness input necessary for SlabGenerator to produce slab with correct number of layers
    """
    x0,dx,n = 2.,0.5,0. # Angstrom


    def get_slabs(x: float) -> pmg.Structure:
                slab_gen = SlabGenerator(initial_structure     = unit
                                            ,miller_index      = facet
                                            ,min_slab_size     = x
                                            ,min_vacuum_size   = vac
                                            ,center_slab       = True
                                            ,primitive         = True
                                            ,lll_reduce        = True)
                return slab_gen.get_slabs(symmetrize=sym,repair=True)

    while n<num_atoms_per_layer*layers:
        x0+=dx
        slab_strucs = get_slabs(x0)
        if len(slab_strucs) == 0:
            n = 0
        else:
            n = len(slab_strucs[0])
        if n > 100:
            raise ValueError('over 50 atoms in primitive surface column (' +str(len(get_slabs(x0)))+') not compatible with requested # of layers ('+str(layers)+')')
        if len(slab_strucs)>1:
            print('WARNING MULTIPLE SLABS FOR FACE = {}'.format(facet))
    atoms_objs = []
    for struc in slab_strucs:
        atoms_obj = AseAtomsAdaptor.get_atoms(struc)
        atoms_obj = reindex_by_height(atoms_obj)
        while len(atoms_obj) != layers*num_atoms_per_layer:
            for ind in range(num_atoms_per_layer):
                del atoms_obj[-1]
        atoms_objs.append(atoms_obj)
    return atoms_objs

def reindex_by_height(atoms : ase.Atoms) -> ase.Atoms:
    H = [atom.position[2] for atom in atoms]
    atoms_new = atoms.copy()
    ind = np.argsort(H)
    del atoms_new[:]
    for i in ind:
        atoms_new.append(atoms[i])
    return atoms_new

def flip_atoms(atoms: ase.Atoms) -> ase.Atoms:
    flipped_atoms = atoms.copy()
    flipped_atoms.rotate(v = [1,0,0], a=180, center = 'COP')
    return flipped_atoms

def align_unit_cell(atoms: ase.Atoms) -> ase.Atoms:
    angle = np.arccos(np.dot(atoms.cell[0],[1,0,0])/np.linalg.norm(atoms.cell[0]))
    angle *= 180/np.pi
    rotated_atoms = atoms.copy()
    rotated_atoms.rotate(v = [1,0,0], a = angle, rotate_cell = True)
    return rotated_atoms


def tag_atoms(atoms                 : ase.Atoms
             ,num_layers            : int
             ,num_bins              : int       = 20
             ) -> ase.Atoms:
    # reindex the input Atoms object by height and tag using evenly placed bins
    # that divide the layers into 20
    #atoms     = AseAtomsAdaptor.get_atoms(atoms)
    atoms_new  = reindex_by_height(atoms)
    num_atoms_per_layer = len(atoms_new)/num_layers
    height     = [round(atom.position[2],4) for atom in atoms_new]
    binH       = np.digitize(height,np.linspace(min(height)*0.99,max(height)*1.01,num_bins))
    binind     = list(np.sort(list(set(binH))))
    binind.reverse()
    tag_curr = num_layers + 1
    for i in range(len(binH)):
        if i%num_atoms_per_layer==0:
            tag_curr -= 1
        atoms_new[i].tag = tag_curr
    return atoms_new

def constrainAtoms(atoms              : ase.Atoms
                  ,constrained_layers : typ.List[int]
                  ,sym                : bool
                  ) -> ase.Atoms:
    # maxTag = max([a.tag for a in atoms])
    # if sym:
    #     if maxTag%2 != nConstrained%2: raise ValueError('Impossible to have symmetric slab with %d layers and %d fixed'%(maxTag,nConstrained))
    #     else:     skip = (maxTag - nConstrained) / 2 #should be integer
    # else: skip = 0
    # constrainedLayers = range(maxTag-nConstrained+1-skip,maxTag+1-skip)
    constrained_atoms = []
    for i,a in enumerate(atoms):
        if a.tag in constrained_layers: constrained_atoms.append(i)
    return constrained_atoms

def is_symmetric_in_z(slab : pmg.Structure) -> bool:
        symmops = SpacegroupAnalyzer(slab).get_symmetry_operations()
        rotation_matrices = [op.rotation_matrix for op in symmops]
        x = np.array([[1.,0.,0.],[0.,-1.,0.],[0.,0.,-1.]])
        return any([(x == rot_mat).all() for rot_mat in rotation_matrices])

def get_unique_facets(atoms     : ase.Atoms
                     ,max_index : int
                     ) -> typ.List[typ.Tuple[int,int,int]]:
    bulkstruct = AseAtomsAdaptor.get_structure(atoms) #Converts ASE atoms to PyMatgen structures (atom -> sites)
    sga = SpacegroupAnalyzer(bulkstruct, 0.1) #Find crystal structure
    bulkstruct = sga.get_conventional_standard_structure()
    return get_symmetrically_distinct_miller_indices(bulkstruct,max_index)

def get_emt_energy(atoms        : ase.Atoms) -> float:
    import emt  # type: ignore
    atoms.set_calculator(emt.EMT())
    return atoms.get_potential_energy()

##############
# Main Scripts
##############

def makeSlab(ase_atoms  : ase.Atoms
            ,facet      : typ.Tuple[int,int,int]
            ,lay        : int
            ,sym        : bool
            ,xy         : typ.Tuple[int,int]
            ,vac        : float
            )->typ.List[pmg.Structure]:
    """
    helpful docstring
    """
    pmg_a    = AseAtomsAdaptor.get_structure(ase_atoms)
    sga      = SpacegroupAnalyzer(pmg_a,symprec=0.1)
    unit     = sga.get_conventional_standard_structure()
    num_atoms_per_layer = len(sga.get_primitive_standard_structure())
    atoms_objs = make_slab(unit,facet,num_atoms_per_layer,vac,lay,sym)
    slabs = [AseAtomsAdaptor.get_structure(atoms_obj) for atoms_obj in atoms_objs]
    slabs = list(map(reorient_z,slabs))
    return slabs

def bulk2surf(bulk              : ase.Atoms
             ,facet             : typ.Tuple[int,int,int]
             ,xy                : typ.Tuple[int,int]
             ,layers            : int
             ,constrained       : typ.List[int]
             ,symmetric         : bool
             ,vacuum            : float
             ,turn_on_magmoms   : bool      = False
             ) -> dict:
    """
    helpful docstring
    """
    magmomInit              = 3
    magElems                = ['Fe','Mn','Cr','Co','Ni']
    bare_slabs              = makeSlab(bulk,facet,layers,symmetric,xy,vacuum)
    bare_ase_objs_tops      = list(map(AseAtomsAdaptor.get_atoms, bare_slabs))
    symm_slab_list          = list(map(is_symmetric_in_z,bare_slabs))
    bare_ase_objs_bottoms   = []
    for bare_ase, symm_slab in zip(bare_ase_objs_tops,symm_slab_list):
        if not symm_slab:
            bare_ase_objs_bottoms.append(flip_atoms(bare_ase))

    output_dict = {'top':bare_ase_objs_tops,'bottom':bare_ase_objs_bottoms}
    for side, atoms_list in output_dict.items():
        for i, atoms_curr in enumerate(atoms_list):
            atoms_curr = tag_atoms(atoms_curr, layers)
            constrain_inds = constrainAtoms(atoms_curr,constrained,symmetric)
            atoms_curr.set_constraint(FixAtoms(indices=constrain_inds))
            if turn_on_magmoms:
                magmoms      = [magmomInit if (magmomInit and e in magElems) else 0 for e in atoms_curr.get_chemical_symbols()]
                atoms_curr.set_initial_magnetic_moments(magmoms)
            atoms_curr.set_pbc((True, True, False))
            atoms_curr.center()
            atoms_curr.cell[2,0] = 0
            atoms_curr.cell[2,1] = 0
            atoms_curr = align_unit_cell(atoms_curr)
            atoms_curr*=[xy[0],xy[1],1]
            atoms_curr.__setattr__('side',side)
            atoms_curr.wrap()
            atoms_list[i] = atoms_curr
    return output_dict



# External Modules
import os,shutil,getpass,time,sys,json

################################################################################
def safeMkdir(newDir): # type: ignore
    try:            os.makedirs(newDir)
    except OSError: pass # may fail due to multiple threads or already existing

def safeCopy(path,name): # type: ignore
    try:            shutil.copyfile(path,name)
    except IOError: pass

def clusterRoot(): # type: ignore
    hostname = os.environ['HOSTNAME'].lower()
    if      'sh'    in hostname or  'gpu-' in hostname:    return '/scratch/users/ksb/share/jobs'
    elif    'su'    in hostname:                           return '/nfs/slac/g/suncatfs/ksb/share/jobs'
    elif    'nid'   in hostname:                           return '/global/cscratch1/sd/krisb/share/jobs'
    else: raise ValueError

################################################################################

def log(**kwargs): # type: ignore
    """
    Execute at the end of all scripts. Copies info to external storage, where it should be added to a database
    """

    ######################################################
    print("Collecting environment data...")
    #-----------------------------------------------------
    public_storage    = clusterRoot()+'/'  # type: ignore

    user              = getpass.getuser()
    timestamp         = time.time()
    working_directory = os.getcwd()
    script            = sys.argv[0].split('/')[-1] # e.g. opt.py

    ######################################################
    print("Agglomerating data to be logged in database...")
    #-----------------------------------------------------
    newDir      = public_storage+'%s/%s/'%(user,str(timestamp).replace('.','')) #name of new directory

    allCols     = ['deleted','user',    'timestamp', 'working_directory','storage_directory','kwargs']
    binds       = [0,         user,  int(timestamp),  working_directory,  newDir,          json.dumps(kwargs)]
    runtime     = dict(zip(allCols,binds))

    #################################################
    print("Creating and Populating copy directory...")
    #------------------------------------------------
    safeMkdir(newDir) # type: ignore

    with open(newDir+'runtime.json','w') as f: f.write(json.dumps(runtime))       # Write runtime info to file

    for root, dirs, files in os.walk(working_directory, topdown=False):
        for name in files:
            originalPath = os.path.join(root, name)
            if  name == script:                           # uniform naming system for 'opt.py' script
                safeCopy(originalPath,newDir+'script.py') # type: ignore
            elif '.err' in name:                          # uniform naming system for error file
                safeCopy(originalPath,newDir+'myjob.err') # type: ignore
            elif '.out' in name:                          # uniform naming system from out file
                safeCopy(originalPath,newDir+'myjob.out') # type: ignore
            elif os.stat(originalPath).st_size < 1e9: safeCopy(originalPath,newDir+name)  # type: ignore

    os.system('chmod -R 755 %s'%newDir)

    return 0



# Typing Modules
import typing as typ

# External Modules
import pickle,json
from decimal import Decimal

# Internal Modules
from catalog.misc.atoms import traj_to_json,json_to_traj,classify_system
from catalog.jobs.jobs import Job
"""
Class for version control of "jobs".

Also store versions in this file


Old preprocessing that might need to be addressed:

        params['spinpol']  = typ.Any([x>0 for x in self.atoms.get_initial_magnetic_moments()])

        ###############################################################
        if params.get('kpts') is None and params.get('kptden') is not None:
            recipcell,kpts = self.atoms.get_reciprocal_cell(),[]
            for i in range(3):
                k = 2 * 3.14 * ((recipcell[i]**2).sum())**(0.5) * params['kptden']
                kpts.append(2 * int(round(k / 2)))

            if   self.kind=='surface':  params['kpts'] = json.dumps(kpts[:2]+[1])
            elif self.kind=='molecule': params['kpts'] = json.dumps([1,1,1])
            else:                       params['kpts'] = json.dumps(kpts)
        ##################################################################

        if 'facet_json' in params.keys(): params['facet']=params.pop('facet_json')
        if 'vibids_json' in params.keys(): params['vibids']=params.pop('vibids_json')
        if 'adsorbates_json' in params.keys(): params['adsorbates']=params.pop('adsorbates_json')
        if params.get('structure')=='hexagonal': params['structure'] = 'hcp'

"""
###################################################
# Constants
#----------


# We want a mapping from VERSION NUMBER to FUNCTION
#   where the function fills in the value (typ.Any type)
#   *given* the rest of the parameters (typ.Dict)
# Alternatively, a single function means valid for all verions
DefaultFunc    = typ.Callable[[typ.Dict],typ.Any]
DefaultHandler = typ.Union[DefaultFunc,typ.Dict[int,DefaultFunc]]

#######################################################
# Helper functions
#-----------------

def err(keyname : str) ->  DefaultFunc:
    """
    Throws an error when called - use as 'default' function for required keys
    """
    def f(d : dict) -> typ.Any:
        raise ValueError("Required key <%s> not found in %s"%(keyname,str(d.keys())))
    return f

def const(x : typ.Any) -> DefaultFunc:
    """
    Turns a value into a function that ignores its input and returns that value
    """
    def f(_ : dict) -> typ.Any: return x
    return f

def onlyif(pred  : typ.Callable[[typ.Dict], bool]
          ,dfunc : DefaultFunc
          ) -> typ.Any:
    """
    Applies default function only if condition is met (otherwise None)
    """
    def f(Dictionary : dict)->typ.Any:
        if pred(Dictionary): return dfunc(Dictionary)
        else:                return None
    return f

def dftcode(*codes : str) -> typ.Callable[[typ.Dict], bool]:
    """
    Helpful docstring
    """
    def f(d : dict) -> bool: return d['dftcode'] in codes
    return f

def single_atom(params : typ.Dict ) -> bool:
    """
    Helpful docstring
    """
    atoms      = json_to_traj(params['inittraj'])
    singleatom = len(atoms) == 1
    mol        = params.get('kind')=='molecule'
    return singleatom and mol

def classify(traj_json : str) -> str:
    """
    Identify whether initial traj of a params typ.Dict is bulk/surf/molecule
    """
    return classify_system(json_to_traj(traj_json))

def get_kwargs(d: typ.Dict)->str:
    """
    Try to extract job name from a Version 0 job
    """
    k = d.get('kwargs')
    if k is None: return ''
    else:
        try:
            kw = json.loads(k)
            n  = kw.get('jobname')
            return n if n else ''
        except:
            return ''

def has_parent(d : typ.Dict) -> bool:
    """
    Check for a job needing to have a parent
    """
    conditions = [d['jobkind'] in ['vcrelax','latticeopt']
                 ,d['jobkind']=='relax' and d['kind']=='molecule']
    return not any(conditions)


def check_init_magmoms(d : typ.Dict) -> int:
    """
    Returns true (1) if typ.Any magmoms in inittraj
    """
    try:
        atom_data = json.loads(d['inittraj'])['atom_data']
    except KeyError:
        atom_data = json.loads(d['inittraj'])['atomdata']
    mags     = [a['magmom'] for a in atom_data]
    return any([x>0 for x in mags])


##############################################################################################################
##############################################################################################################
##############################################################################################################
# Classes
#--------
class Key(object):
    """
    dtype       - a type / list of types that we enforce with `isinstance`
                - if we specify a domain of possible values, then this field is
                    redundant (throw error if both/neither specified)

    description - what the key is used for, typ.Any special considerations

    domain      - an (optional) set of values that we enforce the value falls in

    default     - a function which can initialize the value (given the rest of
                    the typ.Dictionary)

    """
    def __init__(self
                ,name    : str
                ,desc    : str                          = '<no description available yet>'
                ,dtype   : typ.Any                      = None # don't know how to typecheck this one
                ,domain  : typ.Optional[list]           = None
                ,process : typ.Optional[DefaultHandler] = None
                ,default : typ.Optional[DefaultHandler] = None
                ) -> None:

        # Validate input upon construction
        init_err  = 'need to specify either a domain of values or a set of'
        init_err += 'possible types for '+name

        assert (domain is None) ^ (dtype is None), init_err

        # Default handling of inputs
        if isinstance(dtype,type):
            dtype = [dtype]         # box a singleton

        if process is None:
            process = lambda d: d.get(self.name)

        if default is None:
            default = err(name)    # default = throw error

        # Store data in fields
        self.name    = name
        self.dtype   = dtype
        self.desc    = desc
        self.domain  = domain
        self.process = process
        self.default = default

    def _check(self
              ,val : typ.Any
              ) -> typ.Any:
        """
        Enforces domain and type specification for a given value
        """
        err_msg  = "Expected one of %s for input %s" % (self.dtype,self.name)
        err_msg += ", but got %s" % val.__class__
        if self.dtype is not None:
            success  = any([isinstance(val,x) for x in self.dtype])

            if not success:
                raise TypeError(err_msg)

            elif self.domain is not None:
                err_msg2  = "Expected %s to be in %s" % (self.dtype,self.domain)
                err_msg2 += ", but got %s" % val
                success   = val in self.domain
                if not success:
                    raise ValueError(err_msg2)

        return val

    def get_val(self
               ,params  : dict
               ,curr_version : int
               ) -> typ.Any:
        """
        Protocol for inferring a value from an arbitrary params typ.Dictionary
        Handles defaults and applies the validation check
        """

        val = params.get(self.name)

        if val is None: process_func = self.default
        else:           process_func = self.process

        version = params.get('version',0) # before the days of version control

        # print('\tgetting val for',self.name)

        if version > curr_version:
            raise ValueError('current version < dict version')

        elif version < curr_version:

            if isinstance(process_func,typ.Dict):
                try:
                    val = process_func[version](params)
                except KeyError:
                    err_msg = "%s has no default for version %d"%(self.name,version)
                    raise ValueError(err_msg)
            else:
                # no version-dependency of default function
                val = process_func(params)

        return self._check(val)

class Version(object):
    """
    Gives a set of keys a unique ID
    """
    def __init__(self
                ,id       : int
                ,keys     : typ.List[Key]
                ) -> None:

        # Validate input upon construction
        assert len(set([x.name for x in keys]))==len(keys),'DUPLICATE KEYS'

        self.id   = id
        self.keys = keys

    def process_dict(self,dict_input : dict)->"Job":
        dict_output = dict_input
        for key in self.keys:
            val = key.get_val(dict_output, self.id)
            if isinstance(val,Decimal):
                val = float(val)
            dict_output.update({key.name:val})
        return Job(dict_output)


#######################################################
# default functions
#--------------------
def pickletraj_to_jsontraj(d : typ.Dict) -> str:
    """
    Converts old traj format (pickled ASE atoms object) into new format
    """
    try:
        a = pickle.loads(d['inittraj_pckl'].encode())
    except UnicodeDecodeError:
        a = pickle.loads(d['inittraj_pckl'].encode('latin1'), encoding='latin1')
    return traj_to_json(a)




from ase.atoms import Atoms, string2symbols   # type: ignore
from ase.units import kB, kJ, kcal, mol   # type: ignore
import numpy as np   # type: ignore

# Sol57LC
sol57_lp    = ['Li_bcc','Na_bcc','K_bcc','Rb_bcc','Ca_fcc','Sr_fcc','Ba_bcc',
               'V_bcc','Nb_bcc','Ta_bcc','Mo_bcc','W_bcc','Fe_bcc',
               'Rh_fcc','Ir_fcc','Ni_fcc','Pd_fcc','Pt_fcc','Cu_fcc',
               'Ag_fcc','Au_fcc','Al_fcc',
               'C_diamond','Si_diamond','Ge_diamond','Sn_diamond',
               'LiH_b1','LiF_b1','LiCl_b1','NaF_b1','NaCl_b1','MgO_b1','MgS_b1',
               'CaO_b1','TiC_b1','TiN_b1','ZrC_b1','ZrN_b1','VC_b1','VN_b1',
               'NbC_b1','NbN_b1','FeAl_b2','CoAl_b2','NiAl_b2','BN_b3',
               'BP_b3','BAs_b3','AlN_b3','AlP_b3','AlAs_b3','GaN_b3',
               'GaP_b3','GaAs_b3','InP_b3','InAs_b3','SiC_b3']
assert len(sol57_lp) == 57
# Sol58LC = Sol57LC + Pb_fcc
sol58_lp    = ['Li_bcc','Na_bcc','K_bcc','Rb_bcc','Ca_fcc','Sr_fcc','Ba_bcc',
               'V_bcc','Nb_bcc','Ta_bcc','Mo_bcc','W_bcc','Fe_bcc',
               'Rh_fcc','Ir_fcc','Ni_fcc','Pd_fcc','Pt_fcc','Cu_fcc',
               'Ag_fcc','Au_fcc','Al_fcc','Pb_fcc',
               'C_diamond','Si_diamond','Ge_diamond','Sn_diamond',
               'LiH_b1','LiF_b1','LiCl_b1','NaF_b1','NaCl_b1','MgO_b1','MgS_b1',
               'CaO_b1','TiC_b1','TiN_b1','ZrC_b1','ZrN_b1','VC_b1','VN_b1',
               'NbC_b1','NbN_b1','FeAl_b2','CoAl_b2','NiAl_b2','BN_b3',
               'BP_b3','BAs_b3','AlN_b3','AlP_b3','AlAs_b3','GaN_b3',
               'GaP_b3','GaAs_b3','InP_b3','InAs_b3','SiC_b3']
assert len(sol58_lp) == 58
# Sol53Ec
sol53_coh    = ['Li_bcc','Na_bcc','K_bcc','Rb_bcc','Ca_fcc','Sr_fcc','Ba_bcc',
                'V_bcc','Nb_bcc','Ta_bcc','Mo_bcc','W_bcc','Fe_bcc',
                'Rh_fcc','Ir_fcc','Ni_fcc','Pd_fcc','Pt_fcc','Cu_fcc',
                'Ag_fcc','Au_fcc','Al_fcc',
                'C_diamond','Si_diamond','Ge_diamond','Sn_diamond',
                'LiH_b1','LiF_b1','LiCl_b1','NaF_b1','NaCl_b1','MgO_b1','MgS_b1',
                'CaO_b1','TiC_b1','TiN_b1','ZrC_b1','ZrN_b1','VC_b1','VN_b1',
                'NbC_b1','NbN_b1','BN_b3','BP_b3','AlN_b3','AlP_b3','AlAs_b3',
                'GaN_b3','GaP_b3','GaAs_b3','InP_b3','InAs_b3','SiC_b3']
assert len(sol53_coh) == 53
# Sol54Ec
sol54_coh    = ['Li_bcc','Na_bcc','K_bcc','Rb_bcc','Ca_fcc','Sr_fcc','Ba_bcc',
                'V_bcc','Nb_bcc','Ta_bcc','Mo_bcc','W_bcc','Fe_bcc',
                'Rh_fcc','Ir_fcc','Ni_fcc','Pd_fcc','Pt_fcc','Cu_fcc',
                'Ag_fcc','Au_fcc','Al_fcc','Pb_fcc',
                'C_diamond','Si_diamond','Ge_diamond','Sn_diamond',
                'LiH_b1','LiF_b1','LiCl_b1','NaF_b1','NaCl_b1','MgO_b1','MgS_b1',
                'CaO_b1','TiC_b1','TiN_b1','ZrC_b1','ZrN_b1','VC_b1','VN_b1',
                'NbC_b1','NbN_b1','BN_b3','BP_b3','AlN_b3','AlP_b3','AlAs_b3',
                'GaN_b3','GaP_b3','GaAs_b3','InP_b3','InAs_b3','SiC_b3']
assert len(sol54_coh) == 54
# BM32
bm32        = ['Li_bcc','Na_bcc','Ca_fcc','Sr_fcc','Ba_bcc',
               'Rh_fcc','Pd_fcc','Cu_fcc','Ag_fcc','Al_fcc',
               'C_diamond','Si_diamond','Ge_diamond','Sn_diamond',
               'LiH_b1','LiF_b1','LiCl_b1','NaF_b1','NaCl_b1','MgO_b1',
               'BN_b3','BP_b3','BAs_b3','AlN_b3','AlP_b3','AlAs_b3',
               'GaN_b3','GaP_b3','GaAs_b3','InP_b3','InAs_b3','SiC_b3']
assert len(bm32) == 32
# hcp
hcp_solids_10 = ['Cd_hcp', 'Co_hcp', 'Os_hcp', 'Ru_hcp', 'Zn_hcp',
                 'Zr_hcp', 'Sc_hcp', 'Be_hcp', 'Mg_hcp', 'Ti_hcp']
assert len(hcp_solids_10) == 10
# SE30
se21 = ['Li110','Na110','K110','Rb110','Ba110','Ca111','Sr111',
        'Nb110','Ta110','Mo110','W110','Fe110','Al111','Ni111',
        'Cu111','Rh111','Pd111','Ag111','Ir111','Pt111','Au111']
se9  = ['Mg0001','Zn0001','Cd0001','Sc0001','Ti0001',
        'Co0001','Zr0001','Ru0001','Os0001']
se30 = se21+se9
assert len(se30) == 30
#
#
#
#==========================================================
# SURFACE ENERGIES
#==========================================================
# see Vitos, Ruban, Skriver, Kollar, Surf Sci 411 (1998) 186-202
# experimental surface energies in J/m**2
#
#
#
surf_data = {
# (110) facets of bcc metals
'Li110': {
    'solid': 'Li_bcc',
    'facet': '110',
    'E_surf': [0.522, 0.525]},
'Na110': {
    'solid': 'Na_bcc',
    'facet': '110',
    'E_surf': [0.261, 0.260]},
'K110': {
    'solid': 'K_bcc',
    'facet': '110',
    'E_surf': [0.145, 0.130]},
'Rb110': {
    'solid': 'Rb_bcc',
    'facet': '110',
    'E_surf': [0.117, 0.110]},
'Cs110': {
    'solid': 'Cs_bcc',
    'facet': '110',
    'E_surf': [0.095, 0.095]},
'Ba110': {
    'solid': 'Ba_bcc',
    'facet': '110',
    'E_surf': [0.380, 0.370]},
'V110': {
    'solid': 'V_bcc',
    'facet': '110',
    'E_surf': [2.622, 2.550]},
'Fe110': {
    'solid': 'Fe_bcc',
    'facet': '110',
    'E_surf': [2.417, 2.475]},
'Nb110': {
    'solid': 'Nb_bcc',
    'facet': '110',
    'E_surf': [2.655, 2.700]},
'Ta110': {
    'solid': 'Ta_bcc',
    'facet': '110',
    'E_surf': [2.902, 3.150]},
'Mo110': {
    'solid': 'Mo_bcc',
    'facet': '110',
    'E_surf': [2.907, 3.000]},
'Ta110': {
    'solid': 'Ta_bcc',
    'facet': '110',
    'E_surf': [2.902, 3.150]},
'W110': {
    'solid': 'W_bcc',
    'facet': '110',
    'E_surf': [3.265, 3.675]},
# (0001) facets of hcp metals
'Mg0001': {
    'solid': 'Mg_hcp',
    'facet': '0001',
    'E_surf': [0.785, 0.760]},
'Zn0001': {
    'solid': 'Zn_hcp',
    'facet': '0001',
    'E_surf': [0.993, 0.990]},
'Cd0001': {
    'solid': 'Cd_hcp',
    'facet': '0001',
    'E_surf': [0.762, 0.740]},
'Sc0001': {
    'solid': 'Sc_hcp',
    'facet': '0001',
    'E_surf': [1.275]},
'Ti0001': {
    'solid': 'Ti_hcp',
    'facet': '0001',
    'E_surf': [1.989, 2.100]},
'Co0001': {
    'solid': 'Co_hcp',
    'facet': '0001',
    'E_surf': [2.522, 2.550]},
'Zr0001': {
    'solid': 'Zr_hcp',
    'facet': '0001',
    'E_surf': [1.909, 2.000]},
'Ru0001': {
    'solid': 'Ru_hcp',
    'facet': '0001',
    'E_surf': [3.043, 3.050]},
'La0001': {
    'solid': 'La_hcp',
    'facet': '0001',
    'E_surf': [1.020]},
'Os0001': {
    'solid': 'Os_hcp',
    'facet': '0001',
    'E_surf': [3.439, 3.450]},
# (111) facets of fcc metals
'Ca111': {
    'solid': 'Ca_fcc',
    'facet': '111',
    'E_surf': [0.502, 0.490]},
'Sr111': {
    'solid': 'Sr_fcc',
    'facet': '111',
    'E_surf': [0.419, 0.410]},
'Al111': {
    'solid': 'Al_fcc',
    'facet': '111',
    'E_surf': [1.143, 1.160]},
'Pb111': {
    'solid': 'Pb_fcc',
    'facet': '111',
    'E_surf': [0.593, 0.600]},
'Ni111': {
    'solid': 'Ni_fcc',
    'facet': '111',
    'E_surf': [2.380, 2.450]},
'Cu111': {
    'solid': 'Cu_fcc',
    'facet': '111',
    'E_surf': [1.790, 1.825]},
'Rh111': {
    'solid': 'Rh_fcc',
    'facet': '111',
    'E_surf': [2.659, 2.700]},
'Pb111': {
    'solid': 'Pb_fcc',
    'facet': '111',
    'E_surf': [0.593, 0.600]},
'Pd111': {
    'solid': 'Pd_fcc',
    'facet': '111',
    'E_surf': [2.003, 2.050]},
'Ag111': {
    'solid': 'Ag_fcc',
    'facet': '111',
    'E_surf': [1.246, 1.250]},
'Ir111': {
    'solid': 'Ir_fcc',
    'facet': '111',
    'E_surf': [3.048, 3.000]},
'Pt111': {
    'solid': 'Pt_fcc',
    'facet': '111',
    'E_surf': [2.489, 2.475]},
'Au111': {
    'solid': 'Au_fcc',
    'facet': '111',
    'E_surf': [1.506, 1.500]},
}
#Surface energy functions
def in_surf_data(name):
    if name not in surf_data:
        raise KeyError('System %s not in database.' % name)

def get_esurf_bulk_data(name):
    in_surf_data(name)
    solid = surf_data[name]['solid']
    lp = get_solid_lattice_parameter(solid)
    facet = surf_data[name]['facet']
    struct = get_solid_crystal_structure(solid)
    symb = get_solid_symbols(solid)
    mm = get_solid_magmom(solid)
    if facet is '111':
        assert struct is 'fcc'
    elif facet is '110':
        assert struct is 'bcc'
    elif facet is '0001':
        assert struct is 'hcp'
    else:
        stop
    return solid, symb, facet, struct, lp, mm

def get_surface_area(facet, lp):
    if facet is '111':
        A = lp**2. * np.sqrt(3.)/4.
    elif facet is '110':
        A = lp**2. / np.sqrt(2.)
    elif facet is '0001':
        A = lp**2. * np.sqrt(3.)/2.
    return A

def convert_to_Jm2(A, e):
    A *= 1.e-20 # m**2
    J = kJ/1000.
    e /= (J*A)
    return e

def get_exp_surface_energy(name):
    in_surf_data(name)
    e = surf_data[name]['E_surf']
    return e, np.mean(e)

def getarg(sys, systems):
    k = np.where(systems==sys)[0]
    if len(k) < 1:
        raise KeyError('%s not in keylist' % sys)
    elif len(k) > 1:
        raise KeyError('multiple %s in keylist' % sys)
    assert len(k) == 1
    return k[0]

def get_dft_latpar(solid, xc):
    in_bulk_data(solid)
    from data_handler_045 import Data_Handler   # type: ignore
    data = Data_Handler()
    if solid in sol58_lp:
        db = 'solids_1_lp'
    elif solid in hcp_solids_10:
        db = 'hcp10_lp'
    else:
        raise KeyError('solid %s not found for xc=%s' % (solid,xc))
    a, b, c, d = data.get_accurate_ref_data(db, xc)
    arg = getarg(solid, b)
    return c[arg]
#
#
#
#===============================================================
# Solids_1 lattice constants, cohesive energies, and bulk moduli
#===============================================================
#
#
#
# References:
# 0) CRC, 64th edition, 1983-1984
# 1) Hao, Fang, Sun, Csonka, Philipsen, Perdew, PRB 85, 014111 (2012)
# 2) Schimka, Harl, Kresse, JCP 134, 024116 (2011)
# 3) Kittel, Introduction to Solid State Physics, 8th edition (2005)
# 4) CODATA, http://www.codata.org/resources/databases/key1.html (01/29/2013)
# 5a) Haglund: 3d metals, PRB (1991)
# 5b) Haglund: 4d metals, PRB (1992)
# 5c) Haglund: 5d metals, PRB (1993)
# 6) Sun, Marsman, Csonka, Ruzsinszky, Hao, Kim, Kresse, Perdew, PRB 84, 035117 (2011)
#
# In general: zero-point phonon corrected lattice constants,
#             cohesive energies, and bulk moduli.
#
bulk_data = {
# Cubic pure solids
'Li_bcc': {
    'name': 'Li_bcc',
    'lattice parameter': 3.451, #ref 1
    'pbe_lp': 3.429,            #ref 1 (dft-code=BAND)
    'cohesive energy': 37.70,   #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 344.,  #ref 3
    'bulk modulus': 13.9,       #ref 2 (phonon-corrected)
    'symbols': 'Li',
    'latex name': "Li",
    'structure':'bcc',
    'magmom': None},
'Na_bcc': {
    'name': 'Na_bcc',
    'lattice parameter': 4.207, #ref 1
    'pbe_lp': 4.197,            #ref 1 (dft-code=BAND)
    'cohesive energy': 25.76,   #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 158.,  #ref 3
    'bulk modulus': 7.7,        #ref 2 (phonon-corrected)
    'symbols': 'Na',
    'latex name': "Na",
    'structure':'bcc',
    'magmom': None},
'K_bcc': {
    'name': 'K_bcc',
    'lattice parameter': 5.211, #ref 1
    'pbe_lp': 5.281,            #ref 1 (dft-code=BAND)
    'cohesive energy': 21.48,   #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 91.,   #ref 3
    'bulk modulus': None,
    'symbols': 'K',
    'latex name': "K",
    'structure':'bcc',
    'magmom': None},
'Rb_bcc': {
    'name': 'Rb_bcc',
    'lattice parameter': 5.580, #ref 1
    'pbe_lp': 5.665,            #ref 1 (dft-code=BAND)
    'cohesive energy': 19.64,   #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 56.,   #ref 3
    'bulk modulus': None,
    'symbols': 'Rb',
    'latex name': "Rb",
    'structure':'bcc',
    'magmom': None},
'Ca_fcc': {
    'name': 'Ca_fcc',
    'lattice parameter': 5.555, #ref 1
    'pbe_lp': 5.521,            #ref 1 (dft-code=BAND)
    'cohesive energy': 42.39,   #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 230.,  #ref 3
    'bulk modulus': 18.7,       #ref 6 (phonon-corrected)
    'symbols': 'Ca',
    'latex name': "Ca",
    'structure':'fcc',
    'magmom': None},
'Sr_fcc': {
    'name': 'Sr_fcc',
    'lattice parameter': 6.042, #ref 1
    'pbe_lp': 6.013,            #ref 1 (dft-code=BAND)
    'cohesive energy': 39.3,    #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 147.,  #ref 3
    'bulk modulus': 12.5,       #ref 6 (phonon-corrected)
    'symbols': 'Sr',
    'latex name': "Sr",
    'structure':'fcc',
    'magmom': None},
'Ba_bcc': {
    'name': 'Ba_bcc',
    'lattice parameter': 5.004, #ref 1
    'pbe_lp': 5.022,            #ref 1 (dft-code=BAND)
    'cohesive energy': 43.2,    #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 110.,  #ref 3
    'bulk modulus': 9.4,        #ref 6 (phonon-corrected)
    'symbols': 'Ba',
    'latex name': "Ba",
    'structure':'bcc',
    'magmom': None},
'V_bcc': {
    'name': 'V_bcc',
    'lattice parameter': 3.024, #ref 1
    'pbe_lp': 2.997,            #ref 1 (dft-code=BAND)
    'cohesive energy': 122.12,  #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 380.,  #ref 3
    'bulk modulus': None,
    'symbols': 'V',
    'latex name': "V",
    'structure':'bcc',
    'magmom': None},
'Nb_bcc': {
    'name': 'Nb_bcc',
    'lattice parameter': 3.293, #ref 1
    'pbe_lp': 3.310,            #ref 1 (dft-code=BAND)
    'cohesive energy': 172.758, #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 275.,  #ref 3
    'bulk modulus': None,
    'symbols': 'Nb',
    'latex name': "Nb",
    'structure':'bcc',
    'magmom': None},
'Ta_bcc': {
    'name': 'Ta_bcc',
    'lattice parameter': 3.299, #ref 1
    'pbe_lp': 3.347,            #ref 1 (dft-code=BAND)
    'cohesive energy': 186.765, #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 240.,  #ref 3
    'bulk modulus': None,
    'symbols': 'Ta',
    'latex name': "Ta",
    'structure':'bcc',
    'magmom': None},
'Mo_bcc': {
    'name': 'Mo_bcc',
    'lattice parameter': 3.141, #ref 1
    'pbe_lp': 3.161,            #ref 1 (dft-code=BAND)
    'cohesive energy': 156.92,  #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 450.,  #ref 3
    'bulk modulus': None,
    'symbols': 'Mo',
    'latex name': "Mo",
    'structure':'bcc',
    'magmom': None},
'W_bcc': {
    'name': 'W_bcc',
    'lattice parameter': 3.161, #ref 1
    'pbe_lp': 3.170,            #ref 1 (dft-code=BAND)
    'cohesive energy': 202.70,  #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 400.,  #ref 3
    'bulk modulus': None,
    'symbols': 'W',
    'latex name': "W",
    'structure':'bcc',
    'magmom': None},
'Fe_bcc': {
    'name': 'Fe_bcc',
    'lattice parameter': 2.855, #ref 1
    'pbe_lp': 2.834,            #ref 1 (dft-code=BAND)
    'cohesive energy': 98.94,   #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 470.,  #ref 3
    'bulk modulus': None,
    'symbols': 'Fe',
    'latex name': "Fe",
    'structure':'bcc',
    'magmom': 2.22},
'Rh_fcc': {
    'name': 'Rh_fcc',
    'lattice parameter': 3.793, #ref 1
    'pbe_lp': 3.827,            #ref 1 (dft-code=BAND)
    'cohesive energy': 132.79,   #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 480.,  #ref 3
    'bulk modulus': 272.1,      #ref 2 (phonon-corrected)
    'symbols': 'Rh',
    'latex name': "Rh",
    'structure':'fcc',
    'magmom': None},
'Ir_fcc': {
    'name': 'Ir_fcc',
    'lattice parameter': 3.832, #ref 1
    'pbe_lp': 3.872,            #ref 1 (dft-code=BAND)
    'cohesive energy': 158.78,  #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 420.,  #ref 3
    'bulk modulus': None,
    'symbols': 'Ir',
    'latex name': "Ir",
    'structure':'fcc',
    'magmom': None},
'Ni_fcc': {
    'name': 'Ni_fcc',
    'lattice parameter': 3.509, #ref 1
    'pbe_lp': 3.517,            #ref 1 (dft-code=BAND)
    'cohesive energy': 102.213, #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 450.,  #ref 3
    'bulk modulus': None,
    'symbols': 'Ni',
    'latex name': "Ni",
    'structure':'fcc',
    'magmom': 0.64},
'Pd_fcc': {
    'name': 'Pd_fcc',
    'lattice parameter': 3.876, #ref 1
    'pbe_lp': 3.932,            #ref 1 (dft-code=BAND)
    'cohesive energy': 90.20,   #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 274.,  #ref 3
    'bulk modulus': 198.1,      #ref 2 (phonon-corrected)
    'symbols': 'Pd',
    'latex name': "Pd",
    'structure':'fcc',
    'magmom': None},
'Pt_fcc': {
    'name': 'Pt_fcc',
    'lattice parameter': 3.913, #ref 1
    'pbe_lp': 3.971,            #ref 1 (dft-code=BAND)
    'cohesive energy': 134.9,   #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 240.,  #ref 3
    'bulk modulus': None,
    'symbols': 'Pt',
    'latex name': "Pt",
    'structure':'fcc',
    'magmom': None},
'Cu_fcc': {
    'name': 'Cu_fcc',
    'lattice parameter': 3.595, #ref 1
    'pbe_lp': 3.630,            #ref 1 (dft-code=BAND)
    'cohesive energy': 80.36,   #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 343.,  #ref 3
    'bulk modulus': 145.0,      #ref 2 (phonon-corrected)
    'symbols': 'Cu',
    'latex name': "Cu",
    'structure':'fcc',
    'magmom': None},
'Ag_fcc': {
    'name': 'Ag_fcc',
    'lattice parameter': 4.063, #ref 1
    'pbe_lp': 4.150,            #ref 1 (dft-code=BAND)
    'cohesive energy': 68.0,    #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 225.,  #ref 3
    'bulk modulus': 110.8,      #ref 2 (phonon-corrected)
    'symbols': 'Ag',
    'latex name': "Ag",
    'structure':'fcc',
    'magmom': None},
'Au_fcc': {
    'name': 'Au_fcc',
    'lattice parameter': 4.061, #ref 1
    'pbe_lp': 4.147,            #ref 1 (dft-code=BAND)
    'cohesive energy': 87.46,   #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 165.,  #ref 3
    'bulk modulus': None,
    'symbols': 'Au',
    'latex name': "Au",
    'structure':'fcc',
    'magmom': None},
'Al_fcc': {
    'name': 'Al_fcc',
    'lattice parameter': 4.019, #ref 1
    'pbe_lp': 4.037,            #ref 1 (dft-code=BAND)
    'cohesive energy': 78.3,    #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 428.,  #ref 3
    'bulk modulus': 82.0,       #ref 2 (phonon-corrected)
    'symbols': 'Al',
    'latex name': "Al",
    'structure':'fcc',
    'magmom': None},
'Pb_fcc': {
    'name': 'Pb_fcc',
    'lattice parameter': 4.912, #ref 1
    'pbe_lp': 5.040,            #ref 1 (dft-code=BAND)
    'cohesive energy': 46.81,   #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 105.,  #ref 3
    'bulk modulus': None,
    'symbols': 'Pb',
    'latex name': "Pb",
    'structure':'fcc',
    'magmom': None},
'C_diamond': {
    'name': 'C_diamond',
    'lattice parameter': 3.555, #ref 1
    'pbe_lp': 3.571,            #ref 1 (dft-code=BAND)
    'cohesive energy': 170.0,   #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 2230., #ref 3
    'bulk modulus': 454.7,      #ref 2 (phonon-corrected)
    'symbols': 'C',
    'latex name': "C",
    'structure':'diamond',
    'magmom': None},
'Si_diamond': {
    'name': 'Si_diamond',
    'lattice parameter': 5.422, #ref 1
    'pbe_lp': 5.468,            #ref 1 (dft-code=BAND)
    'cohesive energy': 106.52,  #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 645.,  #ref 3
    'bulk modulus': 100.8,      #ref 2 (phonon-corrected)
    'symbols': 'Si',
    'latex name': "Si",
    'structure':'diamond',
    'magmom': None},
'Ge_diamond': {
    'name': 'Ge_diamond',
    'lattice parameter': 5.644, #ref 1
    'pbe_lp': 5.764,            #ref 1 (dft-code=BAND)
    'cohesive energy': 88.25,   #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 374.,  #ref 3
    'bulk modulus': 77.3,       #ref 2 (phonon-corrected)
    'symbols': 'Ge',
    'latex name': "Ge",
    'structure':'diamond',
    'magmom': None},
'Sn_diamond': {
    'name': 'Sn_diamond',
    'lattice parameter': 6.476, #ref 1
    'pbe_lp': 6.659,            #ref 1 (dft-code=BAND)
    'cohesive energy': 72.01,   #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 200.,  #ref 3
    'bulk modulus': 55.5,       #ref 2 (phonon-corrected)
    'symbols': 'Sn',
    'latex name': "Sn",
    'structure':'diamond',
    'magmom': None},
# Cubic compounds
'LiH_b1': {
    'name': 'LiH_b1',
    'lattice parameter': 3.979, #ref 2
    'pbe_lp': 4.006,            #ref 2 (dft-code=VASP)
    'cohesive energy': 240.,    #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 40.1,       #ref 2 (phonon-corrected)
    'symbols': 'LiH',
    'latex name': "LiH",
    'structure':'rocksalt',
    'magmom': None},
'LiF_b1': {
    'name': 'LiF_b1',
    'lattice parameter': 3.974, #ref 1
    'pbe_lp': 4.064,            #ref 1 (dft-code=BAND)
    'cohesive energy': 430.,    #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 76.3,       #ref 2 (phonon-corrected)
    'symbols': 'LiF',
    'latex name': "LiF",
    'structure':'rocksalt',
    'magmom': None},
'LiCl_b1': {
    'name': 'LiCl_b1',
    'lattice parameter': 5.072, #ref 1
    'pbe_lp': 5.147,            #ref 1 (dft-code=BAND)
    'cohesive energy': 346.,    #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 38.7,       #ref 2 (phonon-corrected)
    'symbols': 'LiCl',
    'latex name': "LiCl",
    'structure':'rocksalt',
    'magmom': None},
'NaF_b1': {
    'name': 'NaF_b1',
    'lattice parameter': 4.570, #ref 1
    'pbe_lp': 4.700,            #ref 1 (dft-code=BAND)
    'cohesive energy': 383.,    #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 53.1,       #ref 2 (phonon-corrected)
    'symbols': 'NaF',
    'latex name': "NaF",
    'structure':'rocksalt',
    'magmom': None},
'NaCl_b1': {
    'name': 'NaCl_b1',
    'lattice parameter': 5.565, #ref 1
    'pbe_lp': 5.695,            #ref 1 (dft-code=BAND)
    'cohesive energy': 322.,    #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 27.6,       #ref 2 (phonon-corrected)
    'symbols': 'NaCl',
    'latex name': "NaCl",
    'structure':'rocksalt',
    'magmom': None},
'MgO_b1': {
    'name': 'MgO_b1',
    'lattice parameter': 4.188, #ref 1
    'pbe_lp': 4.255,            #ref 1 (dft-code=BAND)
    'cohesive energy': 502.,    #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 169.8,       #ref 2 (phonon-corrected)
    'symbols': 'MgO',
    'latex name': "MgO",
    'structure':'rocksalt',
    'magmom': None},
'MgS_b1': {
    'name': 'MgS_b1',
    'lattice parameter': 5.188, #ref 1
    'pbe_lp': 5.228,            #ref 1 (dft-code=BAND)
    'cohesive energy': 91.52,   #ref 0
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 650.,  #ref 1
    'bulk modulus': None,
    'symbols': 'MgS',
    'latex name': "MgS",
    'structure':'rocksalt',
    'magmom': None},
'CaO_b1': {
    'name': 'CaO_b1',
    'lattice parameter': 4.781, #ref 1
    'pbe_lp': 4.832,            #ref 1 (dft-code=BAND)
    'cohesive energy': 126.16,  #ref 4
    'Ecoh unit': 'kcal/mol',
    'Ecoh is phonon-corrected': False,
    'debye temperature': 648.,  #ref 1
    'bulk modulus': None,
    'symbols': 'CaO',
    'latex name': "CaO",
    'structure':'rocksalt',
    'magmom': None},
'TiC_b1': {
    'name': 'TiC_b1',
    'lattice parameter': 4.318, #ref 1
    'pbe_lp': 4.332,            #ref 1 (dft-code=BAND)
    'cohesive energy': 690.5,   #ref 5a
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': None,
    'symbols': 'TiC',
    'latex name': "TiC",
    'structure':'rocksalt',
    'magmom': None},
'TiN_b1': {
    'name': 'TiN_b1',
    'lattice parameter': 4.226, #ref 1
    'pbe_lp': 4.247,            #ref 1 (dft-code=BAND)
    'cohesive energy': 645.9,   #ref 5a
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': None,
    'symbols': 'TiN',
    'latex name': "TiN",
    'structure':'rocksalt',
    'magmom': None},
'ZrC_b1': {
    'name': 'ZrC_b1',
    'lattice parameter': 4.687, #ref 1
    'pbe_lp': 4.708,            #ref 1 (dft-code=BAND)
    'cohesive energy': 765.3,   #ref 5b
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': None,
    'symbols': 'ZrC',
    'latex name': "ZrC",
    'structure':'rocksalt',
    'magmom': None},
'ZrN_b1': {
    'name': 'ZrN_b1',
    'lattice parameter': 4.576, #ref 1
    'pbe_lp': 4.594,            #ref 1 (dft-code=BAND)
    'cohesive energy': 726.0,   #ref 5b
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': None,
    'symbols': 'ZrN',
    'latex name': "ZrN",
    'structure':'rocksalt',
    'magmom': None},
'VC_b1': {
    'name': 'VC_b1',
    'lattice parameter': 4.148, #ref 1
    'pbe_lp': 4.154,            #ref 1 (dft-code=BAND)
    'cohesive energy': 669.5,   #ref 5a
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': None,
    'symbols': 'VC',
    'latex name': "VC",
    'structure':'rocksalt',
    'magmom': None},
'VN_b1': {
    'name': 'VN_b1',
    'lattice parameter': 4.130, #ref 1
    'pbe_lp': 4.116,            #ref 1 (dft-code=BAND)
    'cohesive energy': 602.6,   #ref 5a
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': None,
    'symbols': 'VN',
    'latex name': "VN",
    'structure':'rocksalt',
    'magmom': None},
'NbC_b1': {
    'name': 'NbC_b1',
    'lattice parameter': 4.461, #ref 1
    'pbe_lp': 4.484,            #ref 1 (dft-code=BAND)
    'cohesive energy': 796.8,   #ref 5b
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': None,
    'symbols': 'NbC',
    'latex name': "NbC",
    'structure':'rocksalt',
    'magmom': None},
'NbN_b1': {
    'name': 'NbN_b1',
    'lattice parameter': 4.383, #ref 1
    'pbe_lp': 4.422,            #ref 1 (dft-code=BAND)
    'cohesive energy': 723.3,   #ref 5a
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': None,
    'symbols': 'NbN',
    'latex name': "NbN",
    'structure':'rocksalt',
    'magmom': None},
'FeAl_b2': {
    'name': 'FeAl_b2',
    'lattice parameter': 2.881, #ref 1
    'pbe_lp': 2.868,            #ref 1 (dft-code=BAND)
    'cohesive energy': None,
    'bulk modulus': None,
    'symbols': 'FeAl',
    'latex name': "FeAl",
    'structure':'cesiumchloride',
    'magmom': 0.35},
'CoAl_b2': {
    'name': 'CoAl_b2',
    'lattice parameter': 2.854, #ref 1
    'pbe_lp': 2.851,            #ref 1 (dft-code=BAND)
    'cohesive energy': None,
    'bulk modulus': None,
    'symbols': 'CoAl',
    'latex name': "CoAl",
    'structure':'cesiumchloride',
    'magmom': None},
'NiAl_b2': {
    'name': 'NiAl_b2',
    'lattice parameter': 2.881, #ref 1
    'pbe_lp': 2.892,            #ref 1 (dft-code=BAND)
    'cohesive energy': None,
    'bulk modulus': None,
    'symbols': 'NiAl',
    'latex name': "NiAl",
    'structure':'cesiumchloride',
    'magmom': None},
'BN_b3': {
    'name': 'BN_b3',
    'lattice parameter': 3.594, #ref 1
    'pbe_lp': 3.624,            #ref 1 (dft-code=BAND)
    'cohesive energy': 652.,    #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 410.2,      #ref 2 (phonon-corrected)
    'symbols': 'BN',
    'latex name': "BN",
    'structure':'zincblende',
    'magmom': None},
'BP_b3': {
    'name': 'BP_b3',
    'lattice parameter': 4.527, #ref 1
    'pbe_lp': 4.548,            #ref 1 (dft-code=BAND)
    'cohesive energy': 496.,    #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 168.0,      #ref 2 (phonon-corrected)
    'symbols': 'BP',
    'latex name': "BP",
    'structure':'zincblende',
    'magmom': None},
'BAs_b3': {
    'name': 'BAs_b3',
    'lattice parameter': 4.764, #ref 1
    'pbe_lp': 4.809,            #ref 1 (dft-code=BAND)
    'cohesive energy': None,
    'bulk modulus': 151.1,      #ref 2 (phonon-corrected)
    'symbols': 'BAs',
    'latex name': "BAs",
    'structure':'zincblende',
    'magmom': None},
'AlN_b3': {
    'name': 'AlN_b3',
    'lattice parameter': 4.368, #ref 2
    'pbe_lp': 4.402,            #ref 2 (dft-code=VASP)
    'cohesive energy': 564.0,   #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 206.0,      #ref 2 (phonon-corrected)
    'symbols': 'AlN',
    'latex name': "AlN",
    'structure':'zincblende',
    'magmom': None},
'AlP_b3': {
    'name': 'AlP_b3',
    'lattice parameter': 5.450, #ref 1
    'pbe_lp': 5.504,            #ref 1 (dft-code=BAND)
    'cohesive energy': 417.0,   #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 87.4,       #ref 2 (phonon-corrected)
    'symbols': 'AlP',
    'latex name': "AlP",
    'structure':'zincblende',
    'magmom': None},
'AlAs_b3': {
    'name': 'AlAs_b3',
    'lattice parameter': 5.649, #ref 1
    'pbe_lp': 5.728,            #ref 1 (dft-code=BAND)
    'cohesive energy': 369.0,   #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 75.0,       #ref 2 (phonon-corrected)
    'symbols': 'AlAs',
    'latex name': "AlAs",
    'structure':'zincblende',
    'magmom': None},
'GaN_b3': {
    'name': 'GaN_b3',
    'lattice parameter': 4.523, #ref 1
    'pbe_lp': 4.549,            #ref 1 (dft-code=BAND)
    'cohesive energy': 439.0,   #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 213.7,       #ref 2 (phonon-corrected)
    'symbols': 'GaN',
    'latex name': "GaN",
    'structure':'zincblende',
    'magmom': None},
'GaP_b3': {
    'name': 'GaP_b3',
    'lattice parameter': 5.441, #ref 1
    'pbe_lp': 5.506,            #ref 1 (dft-code=BAND)
    'cohesive energy': 348.0,   #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 89.6,       #ref 2 (phonon-corrected)
    'symbols': 'GaP',
    'latex name': "GaP",
    'structure':'zincblende',
    'magmom': None},
'GaAs_b3': {
    'name': 'GaAs_b3',
    'lattice parameter': 5.641, #ref 1
    'pbe_lp': 5.751,            #ref 1 (dft-code=BAND)
    'cohesive energy': 322.0,   #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 76.7,       #ref 2 (phonon-corrected)
    'symbols': 'GaAs',
    'latex name': "GaAs",
    'structure':'zincblende',
    'magmom': None},
'InP_b3': {
    'name': 'InP_b3',
    'lattice parameter': 5.858, #ref 1
    'pbe_lp': 5.963,            #ref 1 (dft-code=BAND)
    'cohesive energy': 335.0,   #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 72.0,       #ref 2 (phonon-corrected)
    'symbols': 'InP',
    'latex name': "InP",
    'structure':'zincblende',
    'magmom': None},
'InAs_b3': {
    'name': 'InAs_b3',
    'lattice parameter': 6.048, #ref 1
    'pbe_lp': 6.188,            #ref 1 (dft-code=BAND)
    'cohesive energy': 297.0,   #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 58.6,       #ref 2 (phonon-corrected)
    'symbols': 'InAs',
    'latex name': "InAs",
    'structure':'zincblende',
    'magmom': None},
'SiC_b3': {
    'name': 'SiC_b3',
    'lattice parameter': 4.348, #ref 1
    'pbe_lp': 4.378,            #ref 1 (dft-code=BAND)
    'cohesive energy': 625.0,   #ref 2
    'Ecoh unit': 'kJ/mol',
    'Ecoh is phonon-corrected': True,
    'debye temperature': None,
    'bulk modulus': 229.1,      #ref 2 (phonon-corrected)
    'symbols': 'SiC',
    'latex name': "SiC",
    'structure':'zincblende',
    'magmom': None},
# hcp pure solids
'Cd_hcp': {
    'name': "Cd_hcp",
    'lattice parameter': 2.979,
    'lattice parameter 2': 5.620,
    'cohesive energy': 1.16,
    'bulk modulus kittel': 46.7, # GPa
    'symbols': 'Cd',
    'latex name': "Cd",
    'structure':'hcp',
    'magmom': None},
'Co_hcp': {
    'name': "Co_hcp",
    'lattice parameter': 2.507,
    'lattice parameter 2': 4.069,
    'cohesive energy': 4.39,
    'bulk modulus kittel': 191.4, # GPa
    'symbols': 'Co',
    'latex name': "Co",
    'structure':'hcp',
    'magmom': 1.72},
'Os_hcp': {
    'name': "Os_hcp",
    'lattice parameter': 2.734,
    'lattice parameter 2': 4.392,
    'cohesive energy': 8.17,
    'bulk modulus kittel': 418., # GPa
    'symbols': 'Os',
    'latex name': "Os",
    'structure':'hcp',
    'magmom': None},
'Ru_hcp': {
    'name': "Ru_hcp",
    'lattice parameter': 2.706,
    'lattice parameter 2': 4.282,
    'cohesive energy': 6.74,
    'bulk modulus kittel': 320.8, # GPa
    'symbols': 'Ru',
    'latex name': "Ru",
    'structure':'hcp',
    'magmom': None},
'Zn_hcp': {
    'name': "Zn_hcp",
    'lattice parameter': 2.665,
    'lattice parameter 2': 4.947,
    'cohesive energy': 1.35,
    'bulk modulus kittel': 59.8, # GPa
    'symbols': 'Zn',
    'latex name': "Zn",
    'structure':'hcp',
    'magmom': None},
'Ti_hcp': {
    'name': "Ti_hcp",
    'lattice parameter': 2.951,
    'lattice parameter 2': 4.684,
    'cohesive energy': 4.85,
    'bulk modulus kittel': 105.1, # GPa
    'symbols': 'Ti',
    'latex name': "Ti",
    'structure':'hcp',
    'magmom': None},
'Zr_hcp': {
    'name': "Zr_hcp",
    'lattice parameter': 3.232,
    'lattice parameter 2': 5.148,
    'cohesive energy': 6.25,
    'bulk modulus kittel': 83.3, # GPa
    'symbols': 'Zr',
    'latex name': "Zr",
    'structure':'hcp',
    'magmom': None},
'Sc_hcp': {
    'name': "Sc_hcp",
    'lattice parameter': 3.309,
    'lattice parameter 2': 5.268,
    'cohesive energy': 3.90,
    'bulk modulus kittel': 43.5, # GPa
    'symbols': 'Sc',
    'latex name': "Sc",
    'structure':'hcp',
    'magmom': None},
'Be_hcp': {
    'name': "Be_hcp",
    'lattice parameter': 2.286,
    'lattice parameter 2': 3.585,
    'cohesive energy': 3.32,
    'bulk modulus kittel': 100.3, # GPa
    'symbols': 'Be',
    'latex name': "Be",
    'structure':'hcp',
    'magmom': None},
'Mg_hcp': {
    'name': "Mg_hcp",
    'lattice parameter': 3.209,
    'lattice parameter 2': 5.211,
    'cohesive energy': 1.51,
    'bulk modulus kittel': 35.4, # GPa
    'symbols': 'Mg',
    'latex name': "Mg",
    'structure':'hcp',
    'magmom': None},
}
# bulk solid functions
def in_bulk_data(name):
    if name not in bulk_data:
        raise KeyError('System %s not in database.' % name)

def get_solid_lattice_parameter(name):
    in_bulk_data(name)
    return bulk_data[name]['lattice parameter']

def get_hcp_covera(name):
    in_bulk_data(name)
    if name in hcp_solids_10:
        a = bulk_data[name]['lattice parameter']
        c = bulk_data[name]['lattice parameter 2']
        return c/a
    else:
        return 0.

def get_solid_pbe_lp(name):
    in_bulk_data(name)
    if name in sol58_lp:
        return bulk_data[name]['pbe_lp']
    else:
        return None

def get_solid_cohesive_energy(name):
    #in_bulk_data(name)
    #assert name in sol58_lp
    d = bulk_data[name]
    e = d['cohesive energy']
    if e == None:
        return e
    else:
        # units
        unit = d['Ecoh unit']
        if unit == 'kJ/mol':
            e *= kJ/mol
        elif unit == 'kcal/mol':
            e *= kcal/mol
        else:
            stop
        corrected = d['Ecoh is phonon-corrected']
        if corrected == False:
            theta = d['debye temperature']
            assert theta != None
            e += (9./8.)*kB*theta
        return e

def get_solid_bulk_modulus(name):
    in_bulk_data(name)
    if name in bm32:
        return bulk_data[name]['bulk modulus']
    else:
        return None

def get_solid_magmom(name):
    in_bulk_data(name)
    return bulk_data[name]['magmom']

def get_solid_crystal_structure(name):
    in_bulk_data(name)
    struct = bulk_data[name]['structure']
    assert struct in ['fcc', 'bcc', 'diamond', 'hcp',
                      'rocksalt', 'cesiumchloride', 'zincblende']
    return struct

def get_solid_symbols(name):
    in_bulk_data(name)
    return bulk_data[name]['symbols']

def get_solids_latex_name(name):
    in_bulk_data(name)
    d = bulk_data[name]
    struct = d['structure']
    if struct == 'diamond':
        struct = 'dia'
    return d['latex name'] + ' ('+struct+')'

def get_solids_common_name(name):
    return get_solids_latex_name(name)

def setup_bulk(solid, offset=0., set_lp=True):
    from ase.lattice import bulk   # type: ignore
    in_bulk_data(solid)
    symbol = get_solid_symbols(solid)
    s = get_solid_crystal_structure(solid)
    if set_lp:
        lp = offset
    else:
        lp = get_solid_lattice_parameter(solid) + offset
    m = get_solid_magmom(solid)
    if s == 'hcp':
        cov = get_hcp_covera(solid)
        atoms = bulk(symbol, s, a=lp, covera=cov)
    else:
        atoms = bulk(symbol, s, a=lp)
    atoms.set_pbc(True)
    if m != None:
        mm = np.zeros(len(atoms))
        mm[:] = m
        atoms.set_initial_magnetic_moments(mm)
    return atoms

def setup_atom(symbol, vac=3.0, t=0.0, odd=0.0):
    from ase import Atoms   # type: ignore
    from ase.data import chemical_symbols, atomic_numbers, ground_state_magnetic_moments   # type: ignore
    assert symbol in chemical_symbols
    Z = atomic_numbers[symbol]
    m = ground_state_magnetic_moments[Z]
    atoms = Atoms(symbol)
    atoms.center(vacuum=vac)
    atoms.translate([t,t,t])
    atoms.cell[1][1] += odd
    atoms.cell[2][2] += 2*odd
    atoms.set_initial_magnetic_moments([m])
    return atoms

######################
# atomic constituents
######################
atoms_ = np.array([])
for aa in sol53_coh:
    symbs = get_solid_symbols(aa)
    atoms_ = np.append(atoms_, string2symbols(symbs))
sol53_coh_atoms = np.unique(atoms_)
atoms_ = np.array([])
for aa in sol54_coh:
    symbs = get_solid_symbols(aa)
    atoms_ = np.append(atoms_, string2symbols(symbs))
sol54_coh_atoms = np.unique(atoms_)
atoms_ = np.array([])
for aa in sol58_lp:
    symbs = get_solid_symbols(aa)
    atoms_ = np.append(atoms_, string2symbols(symbs))
sol58_atoms = np.unique(atoms_)
del atoms_, aa, symbs

######################
# complete set of solids
######################
solids_tot = np.append(sol58_lp, sol54_coh)
solids_tot = np.append(solids_tot, bm32)
solids_tot = np.unique(solids_tot)
np.unique(sol58_lp)
assert len(solids_tot) == len(np.unique(sol58_lp)) == 58
solids_tot_atoms = np.unique(sol54_coh_atoms)
assert len(solids_tot_atoms) == len(sol54_coh_atoms)



import json
# Internal Modules
from catalog.jobs.version_utils import Version,Key,const,onlyif,pickletraj_to_jsontraj,dftcode,err,has_parent,get_kwargs,classify,single_atom,check_init_magmoms
from decimal import Decimal
################################################
# Constants
#----------
NoneType = type(None)
#######################################################
# data
#-----
keys_v1 = [Key('jobkind'
            ,'What kind of calculation the job is performing'
            ,domain = ['latticeopt,relax,vcrelax,neb,bulkmod,dos,xc,chargemol,fbl']
            )

          ,Key('pw'
            ,'Planewave cutoff, eV'
            ,int)

          ,Key('dftcode'
            ,'Which code is used to generate results'
            ,domain = ['vasp','gpaw','quantumespresso'])

          ,Key('xc'
            ,'Exchange correlation functional'
            ,domain = ['PBE','RPBE','BEEF','mBEEF'])

          ,Key('psp'
            ,'Name of pseudopotential used'
            ,domain = ['sg15','gbrv15pbe','paw','oldpaw'])

          ,Key('kpts'
            ,'List/Tuple of K points (x,y,z)'
            ,[list,tuple]
            ,process = {0:lambda d: json.loads(d['kpts']) if isinstance(d['kpts'],str) else d['kpts']}
            )

          ,Key('spinpol'
            ,'Whether or not calculation is spin polarized'
            ,int
            ,default = check_init_magmoms)


          ,Key('inittraj'
            ,'JSON version of an ASE trajectory file'
            ,str
            ,default = {0 : pickletraj_to_jsontraj})

          ,Key('econv'
            ,'Differential energy for electronic convergence'
            ,[float,Decimal]
            ,default = const(1e-5))

          ,Key('fmax'
            ,'Max force for ionic convergence'
            ,[float,Decimal]
            ,default = const(0.05))

          ,Key('mixing'
            ,'Linear mixing parameter for electronic density'
            ,[float,Decimal]
            ,default = const(0.1))

          ,Key('maxstep'
            ,'Max number of electronic steps'
            ,int
            ,default = const(250))

          ,Key('sigma'
            ,'Fermi temperature'
            ,[float,Decimal]
            ,process = lambda d: float(d['sigma'])
            ,default = const(0.1))

          ,Key('nbands'
            ,'Number of bands (negative means additional empty bands)'
            ,[int,NoneType]
            ,default = onlyif(dftcode('gpaw','quantumespresso')
                             ,const(-12)))

          ,Key('nmix'
            ,'Number of previous steps mixed in the electronic density'
            ,[int,NoneType]
            ,default = onlyif(dftcode('gpaw','quantumespresso')
                             ,const(10)))

          ,Key('hund'
            ,'Flag for managing spin for single-atom calculations in GPAW'
            ,domain  = [0,1,None]
            ,default = onlyif(dftcode('gpaw'),single_atom))

          ,Key('dw'
            ,'Electronic density granularity cutoff'
            ,[int,NoneType]
            ,default = onlyif(dftcode('quantumespresso')
                             ,lambda d: d['pw']*10))

          ,Key('mixingtype'
            ,'Mixing scheme'
            ,domain  = ['plain','local-TF',None]
            ,default = onlyif(dftcode('quantumespresso')
                             ,const('plain')))

          ,Key('gga'
            ,'XC functional (is this redundant given the xc key?) '
            ,domain  = ['PE','BF',None]
            ,default = onlyif(dftcode('vasp')
                             ,const('PE')))

          ,Key('luse_vdw'
            ,'Turn on vdW correction'
            ,domain=[0,1,None]
            ,default = onlyif(dftcode('vasp')
                             ,const(0)))

          ,Key('zab_vdw'
            ,'???'
            ,domain  = [0,1,None]
            ,default = onlyif(lambda p: p.get('gga')=='BF'
                             ,const(-1.8867)))

          ,Key('gamma'
              ,'Include gamma point?'
              ,domain  = [0,1,None]
              ,default = onlyif(dftcode('vasp')
                            ,const(1)))

          ,Key('ismear'
              ,'ISMEAR determines how the partial occupancies are set for each wavefunction (-1 = Fermi, 0 = Gaussian) IS THIS THE SAME THING AS MIXINGTYPE?'
              ,domain  = [-1,0,None]
              ,default = onlyif(dftcode('vasp')
                            ,const(0)))

          ,Key('kind'
            ,'override automatic detection of bulk/surface/molecule'
            ,domain  = ['bulk','surface','molecule']
            ,default = lambda d: classify(d['inittraj']))

          ,Key('dipol'
            ,"JSON'd list of dipole moment center coordinates (initial guesss)"
            ,[str,NoneType]
            ,default = const(None))

          ,Key('algo'
              ,'specify the electronic minimisation algorithm'
              ,domain  = ['Normal','All','Fast','Damped','Diag',None]
              ,default = onlyif(dftcode('vasp')
                               ,const('normal')))

          ,Key('ibrion'
            ,"SEE VASP DOCS"
            ,[int,NoneType]
            ,default = onlyif(dftcode('vasp')
                             ,const(2)))

          ,Key('nelmdl'
            ,"# of non self consistent electronic steps"
            ,[int,NoneType]
            ,default = onlyif(dftcode('vasp')
                             ,const(-5)))

          ,Key('prec'
            ,'influences the default for four sets of parameters (ENCUT; NGX, NGY, NGZ; NGXF, NGYF, NGZF and ROPT)'
            ,domain  = ['Low','Medium','High','Normal','Accurate','Single',None]
            ,default = onlyif(dftcode('vasp')
                             ,const('normal')))


          ,Key('ionic_steps'
            ,'Max # of ionic steps'
            ,int
            ,default = const(250))

          ,Key('lreal'
            ,'Determines whether the projection operators are evaluated in real-space or in reciprocal space'
            ,domain  = ['.TRUE.','.FALSE.','Auto',None]
            ,default = onlyif(dftcode('vasp')
                            ,const('auto')))

          ,Key('lvhar'
            ,'Determines whether the total local potential (file LOCPOT) contains the entire local potential (ionic plus Hartree plus exchange correlation) or the electrostatic contributions only (ionic plus Hartree)'
            ,[int,NoneType]
            ,default = onlyif(dftcode('vasp')
                            ,const(0)))

          ,Key('structure'
            ,"""The meaning of this isn't terribly clear
                - (should an fcc lattice with an interstitial H be classified as fcc?)
                - Functionally, it is only important when doing latticeopt
                -  (cell information is condensed to minimum-length vector
                -  (e.g. trigonal cells are parameterized by a singleton)
                - which is operated upon by scipy.fmin)"""

            ,domain = ['fcc','hcp','bcc','diamond','cscl'
                      ,'rocksalt-unit','zincblende-unit','M3N-unit'
                      ,'rocksalt'     ,'zincblende'     ,'M3N'
                      ,'Li3N', 'antibixbyite'
                      ,'triclinic','2DMaterial',None]

            ,default = onlyif(lambda d: d['kind'] in ['bulk'] and d['jobkind'] == 'latticeopt'
                             ,err('structure')))

          ,Key('facet'
            ,'json-dumped list of three integers (sorry, HCP)'
            ,[str,NoneType]
            ,default = onlyif(lambda d: d['kind']=='surface'
                             ,err('facet')))

          ,Key('adsorbates'
            ,"JSON'd list of lists (each sublist is a list of indices for a particular adsorbate)"
            ,[str,NoneType]
            ,default = const(None))

          ,Key('xtol'
            ,'Convergence threshold for lattice parameters (or angles, in radians) when minimizing energy'
            ,[float,Decimal,NoneType]
            ,default = onlyif(lambda d: d['jobkind']=='latticeopt'
                             ,const(0.005)))

          ,Key('strain'
            ,'The percent (in both +/- directions) by which all lattice vectors will be stained to generate a dE/dV curve'
            ,[float,Decimal,NoneType]
            ,default = onlyif(lambda d: d['jobkind']=='bulkmod'
                             ,const(0.03)))

          ,Key('vibids'
            ,"JSON'd list of indices to vibrate"
            ,[str,NoneType]
            ,default = const(None))

          ,Key('delta'
            ,'Distance by which atoms will be varied to estimate local PES curvature'
            ,[float,Decimal,NoneType]
            ,default = onlyif(lambda d: d['jobkind']=='vib'
                             ,const(0.04)))

          ,Key('bonded_inds'
            ,'json-dumped list of two indices which will be brought together in FBL'
            ,[str,NoneType]
            ,default = onlyif(lambda d: d['jobkind']=='fbl'
                             ,err('bonded_inds')))

          ,Key('energy_cut_off'
            ,'used as a stopping criterion for FBL'
            ,[float,Decimal,int,NoneType]
            ,default = onlyif(lambda d: d['jobkind']=='fbl'
                             ,err('energy_cut_off')))

          ,Key('step_size'
            ,'Angstrom'
            ,[float,Decimal,NoneType]
            ,default = onlyif(lambda d: d['jobkind']=='fbl'
                             ,err('step_size')))

          ,Key('finaltraj'
            ,'???'
            ,[str,NoneType]
            ,default = onlyif(lambda d: d['jobkind']=='neb'
                             ,err('finaltraj')))

          ,Key('k'
            ,'spring constant'
            ,[float,Decimal,NoneType]
            ,default = onlyif(lambda d: d['jobkind']=='neb'
                             ,const(0.3)))

          ,Key('images'
            ,'number of images in NEB'
            ,[int,NoneType]
            ,default = onlyif(lambda d: d['jobkind']=='neb'
                             ,const(5)))

          ,Key('dipole'
            ,"Dipole correction used"
            ,domain=[0,1]
            ,default = lambda d: int(d['kind']=='surface'))

          ,Key('job_name'
            ,'(This used to be in keyword args, which has now been deprecated) '
            ,[str,NoneType]
            ,default = {0 : get_kwargs})

         ,Key('parent'
            ,'Job which the current job is based on (not defined for latticeopt/molecule-relax,vc-relax)'
            ,[str,NoneType]
            ,default = onlyif(has_parent,err('parent')))
         ,Key('cell_dofree'
             ,'Determines which cell parameters vcrelax will try to optimize. (2Dxy will only optimize x and y dimensions)'
             ,domain = ['all','x','y','z','xy','xz','yz','xyz','shape','volume','2Dxy','2Dshape',None]
             ,default = onlyif(lambda d: d['jobkind']=='vcrelax',const(None)))
         ,Key('cell_factor'
             ,'Parameter for VC relax jobs'
             ,[float,Decimal,NoneType]
             ,default = onlyif(lambda d: d['jobkind']=='vcrelax',const(2.)))
            ]

version1 = Version(1,keys_v1)

# keys_v2 can be defined relative to version 1
versions = [version1] # ,version2,...]
current_version = version1
